/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#ifndef IMGh
#define IMGh

#define IMG_penalty 0
#define IMG_image 1
#define IMG_file 2

#define IMG_sobelhplus 1001
#define IMG_sobelvplus 1002
#define IMG_sobeldplus 1003
#define IMG_sobelh 1004
#define IMG_sobelv 1005
#define IMG_sobeld 1006
#define IMG_sobelhjaehne 1007
#define IMG_sobelvjaehne 1008
#define IMG_sobeldjaehne 1009
#define IMG_laplace 1010
#define IMG_position_x 1011
#define IMG_position_y 1012

#define IMG_nullimg 2000

#define IMG_VERTICAL 1
#define IMG_HORIZONTAL 2
#define IMG_ROTATE 4

typedef struct _IMG
{
  int x_size;
  int y_size;
  double **data_list;
  int *data_length;
  int *dis_mode;
  double *dis_scale;
  int warprange;
  int beamsize;
  int class;
  double dis_limit;
  char *name;
}
IMG;

typedef struct _IMG_list
{
  IMG *img;
  int total_nr;
  char *name;
}
IMG_list;

typedef struct _IMG_filter
{
  char *name;
  int x_size;
  int y_size;
  double* data;
  int mode;
}
IMG_filter;

int IMG_get_depth(IMG*);
int IMG_get_depth_list(IMG_list*);
IMG_list* IMG_load_list (char *);
IMG_list* IMG_load_list_ejf (char*);
int IMG_save_list (IMG_list*,char *);
int IMG_copy (IMG, IMG *);
IMG_list* IMG_copy_list (IMG_list *);
IMG_list* IMG_create_new_list(int, int);
IMG_list* IMG_create_empty_list (IMG_list *, int);
int IMG_insert_in_list (IMG_list *, IMG_list *, int);
int IMG_move_in_list(IMG_list*, int, IMG_list*,int);
int IMG_add_dataspace_in_list(IMG_list*, int, int);
int IMG_free_list (IMG_list *);
int IMG_free (IMG *);
int IMG_list_set_warprange (IMG_list *, int);
int IMG_list_set_beamsize (IMG_list *, int);
int IMG_list_set_dis_limit (IMG_list *, double);
int IMG_list_set_dis_mode (IMG_list *, int, int);
int IMG_list_set_dis_scale (IMG_list *, int, double);
int IMG_get_min_max_list (IMG_list *, int, double *, double *);
int IMG_get_max_dim_list(IMG_list*);
int IMG_norm_list (IMG_list *, int, double, double, double, double);
int IMG_set_data_in_list (IMG_list*, int, double);
int IMG_add_border (IMG *, int, int, int, int, double);
int IMG_add_border_list (IMG_list *, int, int, int, int, double);
IMG_list *IMG_shift3x3_list (IMG_list *, double);
int IMG_flip (IMG *, int);
int IMG_flip_list (IMG_list *, int);
int IMG_filter_img (IMG *, int);
int IMG_filter_list (IMG_list *, int);
int IMG_print_list_info (IMG_list *);
int IMG_read_list (char *, int *, int, int **, double ***);
IMG_list* IMG_create_mean_img_list(IMG_list*,double W2D_function(IMG,IMG,int*),int);

int IMG_calculate_filter (double **, int, int, double ***, double ***);
double IMG_calculate_variance (double **, int *, int, int);
int IMG_image_nr_array_print(int*);
int IMG_image_nr_array_sort(int*);
int* IMG_read_number_array (char*);
IMG_list* IMG_img_list_from_image_nr_array(IMG_list*,int*);

void IMG_add_position_x_img(IMG *);
int IMG_add_position_x_list(IMG_list *);
void IMG_add_position_y_img(IMG *);
int IMG_add_position_y_list(IMG_list *);
#endif
