/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <argp.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <IMG.h>
#include "config.h"

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char* mnist_label_filename;
  char* mnist_data_filename;
  char* jf_filename;
  int verbose;              /* The -v flag */
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"verbose", 'v', 0, 0, "Verbose output"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'v':
      arguments->verbose = 1;
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 3)
	{
	  argp_usage(state);
	}
      switch(state->arg_num)
	{
	case 0:
	  arguments->mnist_data_filename=arg;
	  break;
	case 1:
	  arguments->mnist_label_filename=arg;
	  break;
	case 2:
	  arguments->jf_filename=arg;
	  break;
	}
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 2)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1] [FILE2] [FILE3]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"idx2jf -- A programm to convert MNIST idx files.\nFILE1 is the idx3 data file, FILE2 is the idx1 label file and FILE3 will be created in jf-format.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};

int SGI2INTEL(char* pointer,int size)
{
  int i;
  int last_i;
  char temp;

  for(i=4;i<=size;i+=4)
    {
      temp=pointer[i-4];
      pointer[i-4]=pointer[i-1];
      pointer[i-1]=temp;
      temp=pointer[i-3];
      pointer[i-3]=pointer[i-2];
      pointer[i-2]=temp;
      last_i=i;
    }
  return(last_i);
}

IMG_list*
IMG_idx_to_img_list (char* data_filename,char* label_filename)
{
  IMG_list* img_list;
  FILE* data_file;
  FILE* label_file;
  int data_total_nr;
  int label_total_nr;
  int magic_data_number;
  int magic_label_number;
  int image_x_size;
  int image_y_size;
  int img_nr;
  int data_pos;
  unsigned char class;
  unsigned char* char_image_buffer;

  if ((data_file = fopen (data_filename, "r")) == NULL)
    {
      fprintf (stderr, "can't open file %s\n", data_filename);
      exit (0);
    }
  if ((label_file = fopen (label_filename, "r")) == NULL)
    {
      fprintf (stderr, "can't open file %s\n", label_filename);
      exit (0);
    }

  if(fread(&magic_data_number,4,1,data_file)!=1 
     || fread(&data_total_nr,4,1,data_file)!=1
     || fread(&image_y_size,4,1,data_file)!=1
     || fread(&image_x_size,4,1,data_file)!=1)
    {
      fprintf (stderr, "can't read file %s\n", data_filename);
      exit (0); 
    }
  SGI2INTEL((char*) &magic_data_number,4);
  SGI2INTEL((char*) &data_total_nr,4);
  SGI2INTEL((char*) &image_x_size,4);
  SGI2INTEL((char*) &image_y_size,4);

  if(fread(&magic_label_number,4,1,label_file)!=1 
     || fread(&label_total_nr,4,1,label_file)!=1)
    {
      fprintf (stderr, "can't read file %s\n", label_filename);
      exit (0); 
    }
  SGI2INTEL((char*) &magic_label_number,4);
  SGI2INTEL((char*) &label_total_nr,4);

  if(magic_label_number!=2049)
    {
      fprintf (stderr, "file %s has not the magic_key: 0x00000801\n", label_filename);
      exit (0); 
    }   
  if(magic_data_number!=2051)
    {
      fprintf (stderr, "file %s has not the magic_key: 0x00000803\n", data_filename);
      exit (0); 
    }

  if(data_total_nr!=label_total_nr)
    {
      fprintf (stderr, "data file %s and label file %s are not of the same size\n", data_filename,label_filename);
      exit (0); 
    }    

  char_image_buffer=(unsigned char*) malloc(image_x_size*image_y_size*sizeof(unsigned char));
  if(char_image_buffer==NULL)
    {
      fprintf (stderr, "not enough memory for image_buffer in mnist_to_img_list\n");
      exit (0); 
    }    

  img_list=IMG_create_new_list(data_total_nr,2);

  for(img_nr=0;img_nr<data_total_nr;img_nr++)
    {
      if(fread(char_image_buffer,sizeof(unsigned char),image_x_size*image_y_size,data_file)
	 !=image_x_size*image_y_size)
	{
	  fprintf (stderr, "unexpected end of file %s\n", data_filename);
	  exit (0); 
	}
      if(fread(&class,sizeof(unsigned char),1,label_file)!=1)
	{
	  fprintf (stderr, "unexpected end of file %s\n", label_filename);
	  exit (0); 
	} 
      img_list->img[img_nr].data_list[0]
	=(double*)malloc(image_x_size*image_y_size*sizeof(double));
      if(img_list->img[img_nr].data_list[0]==NULL)
	{
	  fprintf (stderr, "not enough memory for img_list in mnist_to_img_list\n");
	  exit (0); 
	} 
      for(data_pos=0;data_pos<image_x_size*image_y_size;data_pos++)
	img_list->img[img_nr].data_list[0][data_pos]=(double)char_image_buffer[data_pos];
      img_list->img[img_nr].data_length[0]=image_x_size*image_y_size;
      img_list->img[img_nr].x_size=image_x_size;
      img_list->img[img_nr].y_size=image_y_size;
      img_list->img[img_nr].class=(int) class;
      
    }
  free(char_image_buffer);
  fclose(data_file);
  fclose(label_file);
  return (img_list);
}

int
main (int argc, char *argv[])
{
  struct arguments arguments;
  IMG_list* img_list;

  /* Set argument defaults */
  arguments.verbose = 0;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  img_list = IMG_idx_to_img_list (arguments.mnist_data_filename,
				  arguments.mnist_label_filename);
  
  if(arguments.verbose)   
    fprintf(stderr,"Total of %d images\n",img_list->total_nr);

  IMG_save_list(img_list,arguments.jf_filename);
  IMG_free_list(img_list);

  exit (1);
}
