#include <stdio.h>
#include <argp.h>
#include <float.h>
#include <math.h>
#include "IMG.h"
#include "config.h"

const char *argp_program_version =
VERSION;

const char *argp_program_bug_address =
PACKAGE_BUGREPORT;

/* This structure is used by main to communicate with parse_opt. */
struct arguments
{
  char* jf_filename_source;
  char* image_nr_filename;
  char* class_filename;
  int feature_width;
  int feature_height;
  int feature_distance;
  int verbose;
  int class;
};

/*
   OPTIONS.  Field 1 in ARGP.
   Order of fields: {NAME, KEY, ARG, FLAGS, DOC}.
*/
static struct argp_option options[] =
{
  {"verbose", 'v', 0, 0, "Verbose output"},
  {"imglist", 'I',"[FILE]",0, "Textfile with image numbers"},
  {"class", 'C',"N",0, "Class to seperate"},
  {"classlist", 'K',"[FILE]",0, "Textfile with class numbers"},
  {"distance", 'd',"N",0, "Distance between each feature"},
  {"width", 'w',"N",0, "Featurewidth"},
  {"height", 'h',"N",0, "Featureheight"},
  {0}
};


/*
   PARSER. Field 2 in ARGP.
   Order of parameters: KEY, ARG, STATE.
*/
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'v':
      arguments->verbose=1;
      break;
    case 'I':
      if (strlen(arg)<=0)
	argp_usage(state);
      arguments->image_nr_filename = arg;
      break;
   case 'C':
      if (sscanf (arg, "%d%s", &(arguments->class), arg) != 1 || arguments->class<0)
	  argp_usage(state);
      break;
   case 'w':
      if (sscanf (arg, "%d%s", &(arguments->feature_width), arg) != 1 || arguments->feature_width<1)
	  argp_usage(state);
      break;
   case 'h':
      if (sscanf (arg, "%d%s", &(arguments->feature_height), arg) != 1 || arguments->feature_height<1)
	  argp_usage(state);
      break;
    case 'd':
      if (sscanf (arg, "%d%s", &(arguments->feature_distance), arg) != 1 || arguments->feature_distance<1)
	  argp_usage(state);
      break;
    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
	argp_usage(state);

      switch(state->arg_num)
	{
	case 0:
	  arguments->jf_filename_source = arg;
	  break;
	}
      break;
    case ARGP_KEY_END:
      if (state->arg_num < 1)
	{
	  argp_usage (state);
	}
      break;
    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/*
   ARGS_DOC. Field 3 in ARGP.
   A description of the non-option command-line arguments
     that we accept.
*/
static char args_doc[] = "[FILE1]";

/*
  DOC.  Field 4 in ARGP.
  Program documentation.
*/
static char doc[] =
"jf2feature -- The program reads the FILE1 in jf-format and seperates small images as featurevectors.";

/*
   The ARGP structure itself.
*/
static struct argp argp = {options, parse_opt, args_doc, doc};


int IMG_img_list_seperate_class(IMG_list* img_list,int class)
{
  int img_nr;
  int img_nr_find;
  int new_total_nr;
  IMG img_temp_ptr;

  new_total_nr=0;
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      if(img_list->img[img_nr].class!=class)
	{
	  for (img_nr_find = img_nr+1; img_nr_find < img_list->total_nr; img_nr_find++)
	    {
	      if(img_list->img[img_nr_find].class==class)
		{
		  img_temp_ptr=img_list->img[img_nr_find];
		  img_list->img[img_nr_find]=img_list->img[img_nr];
		  img_list->img[img_nr]=img_temp_ptr;
		  new_total_nr++;
		  img_nr_find=img_list->total_nr;
		}
	    }
	  if(img_list->img[img_nr].class!=class)
	    img_nr=img_list->total_nr;
	}
      else
	new_total_nr++;
    }

  for (img_nr = new_total_nr; img_nr < img_list->total_nr; img_nr++)
    IMG_free (&(img_list->img[img_nr]));

  img_list->total_nr=new_total_nr;
  return(new_total_nr);
}

int IMG_check_array_for_number(int* number_array,int number)
{
  int pos;

  for(pos=0;number_array[pos]>=0;pos++)
    if(number_array[pos]==number)
      return(0);

  return(1);
}

int IMG_img_list_seperate_class_list(IMG_list* img_list,int* class_array)
{
  int img_nr;
  int img_nr_find;
  int new_total_nr;
  IMG img_temp_ptr;

  new_total_nr=0;
  for (img_nr = 0; img_nr < img_list->total_nr; img_nr++)
    {
      if(IMG_check_array_for_number(class_array,img_list->img[img_nr].class))
	{
	  for (img_nr_find = img_nr+1; img_nr_find < img_list->total_nr; img_nr_find++)
	    {
	      if(!IMG_check_array_for_number(class_array,img_list->img[img_nr_find].class))
		{
		  img_temp_ptr=img_list->img[img_nr_find];
		  img_list->img[img_nr_find]=img_list->img[img_nr];
		  img_list->img[img_nr]=img_temp_ptr;
		  new_total_nr++;
		  img_nr_find=img_list->total_nr;
		}
	    }
	  if(IMG_check_array_for_number(class_array,img_list->img[img_nr].class))
	    img_nr=img_list->total_nr;
	}
      else
	new_total_nr++;
    }

  for (img_nr = new_total_nr; img_nr < img_list->total_nr; img_nr++)
    IMG_free (&(img_list->img[img_nr]));

  img_list->total_nr=new_total_nr;
  return(new_total_nr);
}

int IMG_print_img_features(IMG img,int feature_width,int feature_height, int feature_distance)
{
  int x,y,v,w;

  for(y=0;y<=img.y_size-feature_height;y+=feature_distance)
    for(x=0;x<=img.x_size-feature_width;x+=feature_distance)
      {
	for(w=0;w<feature_height;w++)
	  for(v=0;v<feature_width;v++)
	    printf("%f ",img.data_list[0][(img.y_size*(y+w))+x+v]);
	printf("\n");
      }
  return(0);
}


int
main (int argc, char *argv[])
{
  struct arguments arguments;
  IMG_list* img_list_source;
  IMG_list* img_list_destination;
  int* image_nr_array;
  int* class_array;
  int image_nr;

  /* Set argument defaults */
  arguments.jf_filename_source = NULL;
  arguments.image_nr_filename = NULL;
  arguments.class_filename = NULL;
  arguments.verbose = 0;
  arguments.class = -1;
  arguments.feature_height = 3;
  arguments.feature_width = 3;
  arguments.feature_distance = 1;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  img_list_source=IMG_load_list(arguments.jf_filename_source);
  if(arguments.verbose)
    IMG_print_list_info (img_list_source);
 
  if(arguments.image_nr_filename!=NULL)
    {
      image_nr_array = IMG_read_number_array (arguments.image_nr_filename);
      
      IMG_image_nr_array_sort(image_nr_array);

      if(arguments.verbose)
	{
	  fprintf(stdout,"#image numbers to seperate from file:%s\n",arguments.jf_filename_source);
	  IMG_image_nr_array_print(image_nr_array);
	}
      
      img_list_destination=IMG_img_list_from_image_nr_array(img_list_source,image_nr_array);
    }
  else
    img_list_destination=img_list_source;


  if(arguments.class != -1)
    {
      IMG_img_list_seperate_class(img_list_destination,arguments.class);
    }

  if(arguments.class_filename != NULL)
    {
      class_array = IMG_read_number_array (arguments.class_filename);     
      IMG_img_list_seperate_class_list(img_list_destination,class_array);
    }  

  for(image_nr=0;image_nr<img_list_destination->total_nr;image_nr++)
    {
      if(arguments.verbose)
	printf("#image-nr:%d class:%d\n",image_nr,img_list_destination->img[image_nr].class);
      IMG_print_img_features(img_list_destination->img[image_nr],
			     arguments.feature_width,
			     arguments.feature_height,
			     arguments.feature_distance);
    }

  if(arguments.verbose)
     IMG_print_list_info (img_list_destination); 

  if(arguments.image_nr_filename!=NULL)
    free(image_nr_array);

  if(arguments.image_nr_filename!=NULL)
    IMG_free_list(img_list_source);
  IMG_free_list(img_list_destination);
  exit (1);
}
