/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "DIS.h"
#include "config.h"
#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

DISDAT DIS_disdat_init(int discolumns)
{
  DISDAT disdat;
  int disrows=100;
  int i;

  disdat.disrows=disrows;
  disdat.discolumns=discolumns;
  disdat.counter=0;

  disdat.row=(DISELEMENT*) malloc(disrows*sizeof(DISELEMENT));
  if(disdat.row==NULL)
    {
      fprintf(stderr, "not enough memory for DISDATA\n");
      exit(1);
    }
  disdat.row[0].disfield=(double*) malloc(disrows*discolumns*sizeof(double));
  if(disdat.row[0].disfield==NULL)
    {
      fprintf(stderr, "not enough memory for DISDAT\n");
      exit(1);
    } 
  
  for(i=1;i<disrows;i++)
    disdat.row[i].disfield=&(disdat.row[0].disfield[i*discolumns]);

  disdat.class_field_length=10;
  disdat.class_counter=0;
  disdat.class_field=(int*) malloc(sizeof(int)*disdat.class_field_length);
  if(disdat.class_field==NULL)
    {
      fprintf(stderr, "not enough memory for DISDAT\n");
      exit(1);
    }  

  return(disdat);
}

int DIS_disdat_free(DISDAT disdat)
{
  free(disdat.row[0].disfield);
  free(disdat.row);
  free(disdat.class_field);
  free(disdat.variance_field);
  return(1);
}

int DIS_diselement_compare(void const *a,void const *b)
{
  return( ((DISELEMENT*)a)->test_nr - ((DISELEMENT*)b)->test_nr );
}

int DIS_disfield_compare(void const *a,void const *b)
{
  int temp=0;

  if(((DISELEMENT*)a)->disfield[0]<((DISELEMENT*)b)->disfield[0])
    temp=-1;
  if(((DISELEMENT*)a)->disfield[0]>((DISELEMENT*)b)->disfield[0])
    temp=1;

  return(temp);
}

int DIS_classelement_compare(void const *a,void const *b)
{
  return( *((int*)a) - *((int*)b) );
}

DISDAT DIS_disdat_insert_class(DISDAT disdat,int class)
{
  int i;

  for(i=0;i<disdat.class_counter;i++)
    if(disdat.class_field[i]==class)
      return(disdat);

  if(disdat.class_counter<disdat.class_field_length)
    {
      disdat.class_field[disdat.class_counter]=class;
      disdat.class_counter++;
      return(disdat);
    }
  else
    {
      disdat.class_field_length*=2;
      disdat.class_field=(int*) realloc( disdat.class_field,
					 disdat.class_field_length*sizeof(int) );
      
      if(disdat.class_field==NULL)
	{
	  fprintf(stderr, "not enough memory for DISDAT\n");
	  exit(1);
	}  
      disdat.class_field[i]=class;
      disdat.class_counter++;
    }
  return(disdat);
}

DISDAT DIS_disdat_sort(DISDAT disdat)
{
  int i;
  int class;
  qsort(disdat.row,disdat.counter,sizeof(DISELEMENT),DIS_diselement_compare);
  
  disdat.row[disdat.counter-1].test_counter=1;
  for(i=disdat.counter-2;i>=0;i--)
    if(disdat.row[i+1].test_nr!=disdat.row[i].test_nr)
      disdat.row[i].test_counter=1;
    else
      disdat.row[i].test_counter=disdat.row[i+1].test_counter+1;
 
  for(i=0;i<disdat.counter;i++)
    {
      disdat=DIS_disdat_insert_class(disdat,disdat.row[i].test_class);
      disdat=DIS_disdat_insert_class(disdat,disdat.row[i].train_class);
    }
  for(i=0;i<disdat.counter;i++)
    {
      for(class=0;class<disdat.class_counter;class++)
	if(disdat.row[i].test_class==disdat.class_field[class])
	  break;
	   
      if(class==disdat.class_counter)
	  {
	    fprintf(stderr,"$Dis_disdat_sort class_error\n");
	    exit(-1);
	  }
      disdat.row[i].test_class=class;

      for(class=0;class<disdat.class_counter;class++)
	if(disdat.row[i].train_class==disdat.class_field[class])
	  break;
	   
      if(class==disdat.class_counter)
	  {
	    fprintf(stderr,"$Dis_disdat_sort class_error\n");
	    exit(-1);
	  }
      disdat.row[i].train_class=class;
    }
  
  qsort(disdat.class_field,disdat.class_counter,sizeof(int),DIS_classelement_compare);

  disdat.class_matrix=(int*) malloc(disdat.class_counter*disdat.class_counter*
				    sizeof(int));
  if(disdat.class_matrix==NULL)
    {
      fprintf(stderr, "not enough memory for DISDATA\n");
      exit(1);
    }
			
  return(disdat);
}
  
DISDAT DIS_disdat_insert(DISDAT disdat,char* string)
{
  int i;

  if(disdat.counter==disdat.disrows)
    {
      disdat.disrows*=2;
      disdat.row=(DISELEMENT*) realloc( disdat.row,
					disdat.disrows*sizeof(DISELEMENT) );
      if(disdat.row==NULL)
	{
	  fprintf(stderr, "not enough memory for DISDATA\n");
	  exit(1);
	}
      disdat.row[0].disfield=
	(double*) realloc(disdat.row[0].disfield,
			  disdat.disrows*disdat.discolumns*sizeof(double));
      if(disdat.row[0].disfield==NULL)
	{
	  fprintf(stderr, "not enough memory for DISDAT\n");
	  exit(1);
	} 
      for(i=1;i<disdat.disrows;i++)
	disdat.row[i].disfield=
	  &(disdat.row[0].disfield[i*disdat.discolumns]);
    }

  sscanf(string,"%d|%s",&(disdat.row[disdat.counter].test_nr),string);
  sscanf(string,"%d|%s",&(disdat.row[disdat.counter].train_nr),string);
  sscanf(string,"%d|%s",&(disdat.row[disdat.counter].test_class),string);
  sscanf(string,"%d|%s",&(disdat.row[disdat.counter].train_class),string);
  for(i=0;i<disdat.discolumns;i++)
    sscanf(string,"%lf|%s",&(disdat.row[disdat.counter].disfield[i]),string);

  disdat.counter++;
  return(disdat);
}


int DIS_disdat_print(DISDAT disdat)
{
  int i,j;
  for(i=0;i<disdat.counter;i++)
    {
      printf("%d %d %d %d",disdat.row[i].test_nr,disdat.row[i].train_nr,
	     disdat.row[i].test_class,disdat.row[i].train_class);
      for(j=0;j<disdat.discolumns;j++)
	printf(" %g",disdat.row[i].disfield[j]);
      printf("\n");
    }
  for(i=0;i<disdat.class_counter;i++)
    printf("%d ",disdat.class_field[i]);
  printf("-%d-\n",disdat.class_counter);
  return(1);
}

int DIS_disdat_print_best_ER(DISDAT disdat)
{
  int i;
  int test_total_nr=0;
  int test_nr=-1;
  int found; // what should the initial value of found be????
  int right=0;
  
  for(i=0;i<disdat.counter;i++)
    {
      if(test_nr!=disdat.row[i].test_nr)
	{
	  test_total_nr++;
	  test_nr=disdat.row[i].test_nr;
	  found=0;
	}
      if(found==0)
	if(disdat.row[i].test_class==disdat.row[i].train_class)
	  {
	    found=1;
	    right++;
	  }
    }
  printf("Tests:%d\n",test_total_nr);  
  printf("Best possible error rate:%0.2f\%\n",((double) test_total_nr - (double) right)*100/(double) test_nr);

  return(1);
}

int DIS_classifier_KD(DISDAT disdat,int disrow_offset)
{
  int i,k;
  DISELEMENT* temp_data;
  double* temp_disfield;
  double* class_p_sum;
  int* class_sum;
  int best_class;
  double dis_min;
  double exp_dis_min;
  double sqrt_2pi;
  double exp_temp;

  temp_data=(DISELEMENT*) malloc(sizeof(DISELEMENT)*disdat.row[disrow_offset].test_counter);
  if(disdat.row[0].disfield==NULL)
    {
      fprintf(stderr, "not enough memory for kNN temp_data\n");
      exit(1);
    } 
  temp_disfield=(double*) malloc(sizeof(double)*disdat.row[disrow_offset].test_counter);
  class_p_sum=(double*) malloc(sizeof(double)*disdat.class_counter);
  class_sum=(int*) malloc(sizeof(int)*disdat.class_counter);
  if(temp_disfield==NULL || class_p_sum==NULL || class_sum==NULL)
    {
      fprintf(stderr, "not enough memory for kNN temp_data\n");
      exit(1);
    } 
  for(i=0;i<disdat.row[disrow_offset].test_counter;i++)
    {
      temp_data[i]=disdat.row[i+disrow_offset];
      temp_data[i].disfield=&(temp_disfield[i]);
      temp_data[i].disfield[0]=disdat.row[i+disrow_offset].disfield[disdat.disfield_nr];
    }

  qsort(temp_data,disdat.row[disrow_offset].test_counter,sizeof(DISELEMENT),DIS_disfield_compare);
  
  sqrt_2pi=sqrt(2*3.141592654);
  dis_min=temp_data[0].disfield[0]/(disdat.variance_field[temp_data[0].train_class]*2);
  exp_dis_min=exp(dis_min);
  if(exp_dis_min==0)
    {
      fprintf(stderr,"The minimum distance for one test was to large to use the exp() function\n");
    }
  
  for(i=0;i<disdat.class_counter;i++)
    {
      class_p_sum[i]=0.0;
      class_sum[i]=0;
    }
  k=disdat.k;
  if(k>disdat.row[disrow_offset].test_counter)
    k=disdat.row[disrow_offset].test_counter;
  for(i=0;i<k;i++)
    {
      exp_temp=exp(((-0.5)*(temp_data[i].disfield[0]
			    /disdat.variance_field[temp_data[i].train_class]))
		   +dis_min);
      class_p_sum[temp_data[i].train_class]+=exp_temp;
      class_sum[temp_data[i].train_class]++;

      if(exp_temp==0)
	{
	  fprintf(stderr,"A distance was to large to use the exp() function\n");
	}
    }

  for(i=0;i<disdat.class_counter;i++)
    class_p_sum[i]=(class_p_sum[i]*exp_dis_min)/(sqrt(disdat.variance_field[i])*sqrt_2pi);

  best_class=0;
  for(i=1;i<disdat.class_counter;i++)
    {
      if(class_p_sum[i]>class_p_sum[best_class])
	best_class=i;
    }
  
  disdat.class_matrix[(best_class*disdat.class_counter)+disdat.row[disrow_offset].test_class]++;  

  if(disdat.classifier_type & DIS_verbose_out)
    {
      printf("test_nr: %d ",disdat.row[disrow_offset].test_nr);
      if((best_class==disdat.row[disrow_offset].test_class))
	printf("good classified\n");
      else
	printf("wrong classified\n");
      for(i=0;i<disdat.class_counter;i++)
	if(i!=disdat.row[disrow_offset].test_class)
	  printf("%4d ",i);
	else
	  printf("%4d<",i);  
      printf("\n");
      for(i=0;i<disdat.class_counter;i++)
	{
	  printf("%.2f",class_p_sum[i]);
	  if(i!=best_class)
	    printf(" ");
	  else
	    printf("<");
	}
      printf("\n");
    }

  free(class_p_sum);
  free(class_sum);
  free(temp_disfield);
  free(temp_data);
  return((best_class!=disdat.row[disrow_offset].test_class));
}


int DIS_classifier_kNN(DISDAT disdat,int disrow_offset)
{
  int i,k;
  DISELEMENT* temp_data;
  double* temp_disfield;
  double* class_dis_sum;
  int* class_sum;
  int best_class=0;

  temp_data=(DISELEMENT*) malloc(sizeof(DISELEMENT)*disdat.row[disrow_offset].test_counter);
  if(disdat.row[0].disfield==NULL)
    {
      fprintf(stderr, "not enough memory for kNN temp_data\n");
      exit(1);
    } 
  temp_disfield=(double*) malloc(sizeof(double)*disdat.row[disrow_offset].test_counter);
  class_dis_sum=(double*) malloc(sizeof(double)*disdat.class_counter);
  class_sum=(int*) malloc(sizeof(int)*disdat.class_counter);
  if(temp_disfield==NULL || class_dis_sum==NULL || class_sum==NULL)
    {
      fprintf(stderr, "not enough memory for kNN temp_data\n");
      exit(1);
    } 
  for(i=0;i<disdat.row[disrow_offset].test_counter;i++)
    {
      temp_data[i]=disdat.row[i+disrow_offset];
      temp_data[i].disfield=&(temp_disfield[i]);
      temp_data[i].disfield[0]=disdat.row[i+disrow_offset].disfield[disdat.disfield_nr];
    }

  qsort(temp_data,disdat.row[disrow_offset].test_counter,sizeof(DISELEMENT),DIS_disfield_compare);

  for(i=0;i<disdat.class_counter;i++)
    {
      class_dis_sum[i]=0.0;
      class_sum[i]=0;
    }
  k=disdat.k;
  if(k>disdat.row[disrow_offset].test_counter)
    k=disdat.row[disrow_offset].test_counter;
  for(i=0;i<k;i++)
    {
      class_dis_sum[temp_data[i].train_class]+=temp_data[i].disfield[0];
      class_sum[temp_data[i].train_class]++;
    }

  for(i=1;i<disdat.class_counter;i++)
    {
      if(class_sum[i]>class_sum[best_class])
	best_class=i;
      if(class_sum[i]==class_sum[best_class] 
	 && class_dis_sum[i]<class_dis_sum[best_class])
	best_class=i;
    }
  
  disdat.class_matrix[(best_class*disdat.class_counter)+disdat.row[disrow_offset].test_class]++;  

  if(disdat.classifier_type & DIS_verbose_out )
    {
      printf("test_nr: %d ",disdat.row[disrow_offset].test_nr);
      if((best_class==disdat.row[disrow_offset].test_class))
	printf("good classified\n");
      else
	printf("wrong classified\n");
      for(i=0;i<disdat.class_counter;i++)
	if(i!=disdat.row[disrow_offset].test_class)
	  printf("%4d ",i);
	else
	  printf("%4d<",i);  
      printf("\n");
      for(i=0;i<disdat.class_counter;i++)
	{
	  printf("%.2f",((double) class_sum[i])*(1.0/k));
	  if(i!=best_class)
	    printf(" ");
	  else
	    printf("<");
	}
      printf("\n");
    }

  free(class_dis_sum);
  free(class_sum);
  free(temp_disfield);
  free(temp_data);
  return((best_class!=disdat.row[disrow_offset].test_class));
}

int DIS_disdat_classifier(DISDAT disdat)
{
  int test_index;
  int errors;
  int disrow_offset;
  int disfield_start;
  int disfield_end;
  int i,j;

  if(disdat.disfield_nr==-1)
    {
      disfield_start=0;
      disfield_end=disdat.discolumns;
    }
  else
    {
      disfield_start=disdat.disfield_nr;
      disfield_end=disdat.disfield_nr+1;     
    }
  for(disdat.disfield_nr=disfield_start;disdat.disfield_nr<disfield_end;disdat.disfield_nr++)
    {
      disrow_offset=0;
      errors=0;
      for(i=0;i<disdat.class_counter*disdat.class_counter;i++)
	disdat.class_matrix[i]=0;

	for(test_index=0;disrow_offset<disdat.counter;test_index++)
	  {
	    switch(disdat.classifier_type&255)
	      {
	      case DIS_class_kNN:
		errors+=DIS_classifier_kNN(disdat,disrow_offset);
		break;
	      case DIS_class_KD:
		errors+=DIS_classifier_KD(disdat,disrow_offset);
		break;
	      }
	    disrow_offset+=disdat.row[disrow_offset].test_counter;
	  }

	if(disdat.classifier_type & DIS_verbose_matrix)
	  {
	    for(i=0;i<disdat.class_counter;i++)
		printf("%4d ",disdat.class_field[i]);
	    printf("\n");	      
	    for(i=0;i<disdat.class_counter;i++)
	      printf("-----");
	    printf("\n");
	    for(i=0;i<disdat.class_counter;i++)
	      {
		j=0;
		for(disrow_offset=0;disrow_offset<disdat.counter;disrow_offset+=disdat.row[disrow_offset].test_counter)
		  if(disdat.row[disrow_offset].test_class==i)
		    j++;
		printf("%4d ",j);
	      }
	    printf("\n");
	    for(i=0;i<disdat.class_counter;i++)
	      printf("-----");
	    printf("\n");
		
	    for(i=0;i<disdat.class_counter;i++)
	      {
		for(j=0;j<disdat.class_counter;j++)    
		  printf("%4d ",disdat.class_matrix[(j*disdat.class_counter)+i]);
		printf("\n");
	      }
	  }
	printf("#classifier ");
	switch(disdat.classifier_type&255)
	  {
	  case DIS_class_kNN:
	    printf("kNN k=%d ",disdat.k);
	    break;
	  case DIS_class_KD:
	    printf("KD ");
	    break;
	  }
	printf("disfield=%d ",disdat.disfield_nr);
	printf("test_total=%d ",test_index);
	printf("wrong=%d ",errors);
	printf("=> error=%0.2f\n",(double) errors*100.0/ (double) test_index );
    }
  return(0);
}

//int disdat_disvariance(DISDAT disdat)

#define DIS_READ_LINE_LENGTH 400
DISDAT DIS_read_disdat_file(char* disdat_name)
{
#ifdef HAVE_LIBZ
  gzFile gz_file=NULL;
#endif
  FILE* disdat_file=NULL;
  char line_buffer[DIS_READ_LINE_LENGTH];
  int line_nr=0;
  int disrow_nr=0;
  int i,j,sstart,sstop;
  DISDAT disdat;
  int file_end;

  disdat.counter=0;

#ifndef HAVE_LIBZ
  if(strlen(disdat_name)>=3 && strcmp(".gz",&(disdat_name[strlen(disdat_name)-3]))==0) 
    {
      fprintf(stderr,"build with zlib for .gz support\n");
      exit(-1);
    }
#endif
  
  if(!strcmp(disdat_name,"-"))
    disdat_file=stdin;
  else
    {
#ifdef HAVE_LIBZ
      if((gz_file = gzopen (disdat_name, "r")) == NULL)
#else
      if((disdat_file = fopen (disdat_name, "r")) == NULL)
#endif      
	{
          fprintf (stderr, "can't open file %s\n", disdat_name);
          exit (0);
        }
    }


  file_end=0;
  while(!file_end)
    {
      if(!strcmp(disdat_name,"-"))
	{
	  if(line_buffer!=fgets(line_buffer, DIS_READ_LINE_LENGTH, disdat_file))
	    file_end=1;
	}
      else
	{
#ifdef HAVE_LIBZ
	  if(line_buffer!=gzgets (gz_file, line_buffer, DIS_READ_LINE_LENGTH))
	    file_end=1;
#else
	  if(line_buffer!=fgets(line_buffer, DIS_READ_LINE_LENGTH, disdat_file))
	    file_end=1;
#endif
	}
  if(!file_end)
  {
    line_nr++;
    if(strlen(line_buffer)>DIS_READ_LINE_LENGTH-2)
      {
	fprintf (stderr, "line %d exceed max. LINE_LENGTH %d\n",line_nr,DIS_READ_LINE_LENGTH-2);
	exit (0);
      }  


    i=0;
    j=0;
    sstart=0;
    sstop=1;
    while(sstop)
      {
	if(line_buffer[i]=='|')
	  line_buffer[i]=' ';
	if(line_buffer[i]!=' ' && line_buffer[i]!='\t' &&
	   line_buffer[i]!='\n' && line_buffer[i]!='#' &&
	   line_buffer[i]!='$')
	  {
	    if(sstart<i)
	      {
		line_buffer[sstart]=line_buffer[i];
	      }
	    sstart++;
	  }
	else
	  {
	    if(sstart>0 && line_buffer[sstart-1]!='|')
	      {
		j++;
		line_buffer[sstart]='|';
		sstart++;
	      }
	  }
	       
	if(line_buffer[i]=='\0' || line_buffer[i]=='\n' ||
	   line_buffer[i]=='$'  || line_buffer[i]=='#'  )
	  {
	    sstop=0;
	    line_buffer[sstart]='\0';
	  }
	i++;
      }

    if(disrow_nr==0 && j>0)
      {
	disrow_nr=j;
	if(j<5)
	  {
	    fprintf (stderr, "line %d has not the min. line-word-count of 5\n",line_nr);
	    exit (0);	    
	  }	   
	disdat=DIS_disdat_init(disrow_nr-4);
      }
    if(j>0 && disrow_nr!=j)
      {
	fprintf (stderr, "line %d has not the previous line-word-count of %d\n",line_nr,disrow_nr);
	exit (0);	    
      }
    if(j>0)
      //man kann noch prfen,ob test_nr und test_class eindeutig sind
      //und die ersten werte integers sind
	disdat=DIS_disdat_insert(disdat,line_buffer);
  }
    }

  if(disdat_file!=stdin)
#ifdef HAVE_LIBZ
    gzclose(gz_file);
#else
   fclose (disdat_file);
#endif

  if(disdat.counter==0)
    {
      fprintf(stderr,"disdat file %s has no dis lines\n",disdat_name);
      exit(-1);
    }

  disdat=DIS_disdat_sort(disdat);

  disdat.variance_field=(double*) malloc(sizeof(double)*disdat.class_counter);
  if(disdat.variance_field==NULL)
    {
      fprintf(stderr,"Not enough memory for disdat_variance_field\n");
      exit(-1);
    } 
  for(i=0;i<disdat.class_counter;i++)
    disdat.variance_field[i]=1.0;


  fprintf(stdout,"$DIS_read_disdat_file(%s);\n",disdat_name);
  return(disdat);
}

int DIS_disdat_set_variance(DISDAT* disdat,char* var_string,double variance)
{
  int j,i;

  if(strlen(var_string)==0)
    {
      for(i=0;i<disdat->class_counter;i++)
	  disdat->variance_field[i]=1.0;
    }
  else
    {
      i=0;
      while(strlen(var_string)!=0)
	{
	  if(sscanf(var_string,"%lf:%s",&(disdat->variance_field[i]),var_string)<=1)
	    var_string="";
	  i++;
	}
      if(i!=disdat->class_counter)
	{
	  fprintf(stderr,"given variances (%d) not equal with classes (%d)\n",i,disdat->class_counter);
	  exit(-1);
	}
    }

  printf("#v_k*V:");
  for(j=0;j<disdat->class_counter;j++)
    {
      disdat->variance_field[j]*=variance;
      printf(" %g",disdat->variance_field[j]);
    }
  printf("\n");

  return(0);
}

int DIS_disdat_calc(DISDAT disdat)
{
  double* variance;
  double* average;
  int *total_class;
  int j,i;
  int disfield_start;
  int disfield_end;
 
  average= (double*) malloc(sizeof(double)*disdat.class_counter);
  variance= (double*) malloc(sizeof(double)*disdat.class_counter);
  total_class= (int*) malloc(sizeof(int)*disdat.class_counter);
  if(average == NULL || variance == NULL || total_class == NULL)
    {
      fprintf(stderr,"Not enough memory for disdat_calc\n");
      exit(-1);
    }

  if(disdat.disfield_nr==-1)
    {
      disfield_start=0;
      disfield_end=disdat.discolumns;
    }
  else
    {
      disfield_start=disdat.disfield_nr;
      disfield_end=disdat.disfield_nr+1;     
    }
  for(j=disfield_start;j<disfield_end;j++)
    {

      for(i=0;i<disdat.class_counter;i++)
	{
	  average[i]=0;
	  variance[i]=0;
	  total_class[i]=0;
	}
      
      for(i=0;i<disdat.counter;i++)
	{
	  total_class[disdat.row[i].train_class]++;
	  average[disdat.row[i].train_class]+=sqrt(disdat.row[i].disfield[j]);
	  variance[disdat.row[i].train_class]+=disdat.row[i].disfield[j];
	}
      for(i=0;i<disdat.class_counter;i++)
	{
	  average[i]/=total_class[i];
	  variance[i]/=total_class[i];
	  variance[i]-=average[i]*average[i];
	}
           
  
      printf("disfield: %d\n",j);
      printf("Average_k:");
      for(i=0;i<disdat.class_counter;i++)
	printf(" %g",average[i]);
      printf("\n");
      
      printf("Variance_k:");
      for(i=0;i<disdat.class_counter;i++)
	printf(" %g",variance[i]);
      printf("\n");

      printf("Abweichung_k:");
      for(i=0;i<disdat.class_counter;i++)
	printf(" %g",sqrt(variance[i]));
      printf("\n");
    }

  free(average);
  free(variance);
  free(total_class);
  return(0);
}

int DIS_disdat_swap_disfield(DISDAT* disdat,int disfield_nr)
{
  int disrow_nr;
  double double_temp;

  if(disfield_nr>=disdat->discolumns)
    {
      fprintf(stderr,"DIS_disdat_swap_disfield disfield_nr to large\n");
      exit(-1);
    }

  for(disrow_nr=0;disrow_nr<disdat->counter;disrow_nr++)
    {
      double_temp=disdat->row[disrow_nr].disfield[0];
      disdat->row[disrow_nr].disfield[0]=disdat->row[disrow_nr].disfield[disfield_nr];
      disdat->row[disrow_nr].disfield[disfield_nr]=double_temp;
    }

  return(1);
}

int DIS_disdat_sym_distance(DISDAT* disdat)
{
  int disrow_nr;
  int disrow_nr_plus;
  int disfield_nr;
  int disfield_end;
  int test_nr;
  int train_nr;
  int sym;

  disfield_end=disdat->discolumns;

  for(disrow_nr=0;disrow_nr<disdat->counter;disrow_nr++)
    {
      
      test_nr=disdat->row[disrow_nr].test_nr;
      train_nr=disdat->row[disrow_nr].train_nr;   
      sym=0;
      for(disrow_nr_plus=disrow_nr+1;disrow_nr_plus<disdat->counter;disrow_nr_plus++)
	{
	  if(test_nr==disdat->row[disrow_nr_plus].train_nr
	     && train_nr==disdat->row[disrow_nr_plus].test_nr)
	    {
	      for(disfield_nr=0;disfield_nr<disfield_end;disfield_nr++)
		  disdat->row[disrow_nr].disfield[disfield_nr]+=
		    disdat->row[disrow_nr_plus].disfield[disfield_nr];
	      
	      disdat->counter--;
	      disdat->row[disrow_nr_plus]=disdat->row[disdat->counter];
	      
	      sym=1;
	      disrow_nr_plus=disdat->counter;
	    }
	}
      if(sym==0)
	{
	  fprintf(stderr,"DIS_disdat_sym_distance not possible\n");
	  exit(-1);
	}	    
    }

  return(1);
}

DISDAT DIS_disdat_copy(DISDAT disdat_src)
{
  DISDAT disdat;
  int disrow_nr;
  int disfield_nr;
  int class_nr;

  disdat=disdat_src;

  disdat.row=(DISELEMENT*) malloc(disdat.disrows*sizeof(DISELEMENT));
  if(disdat.row==NULL)
    {
      fprintf(stderr, "not enough memory to copy DISDATA\n");
      exit(1);
    }

  for(disrow_nr=0;disrow_nr<disdat.counter;disrow_nr++)
    disdat.row[disrow_nr]=disdat_src.row[disrow_nr];
  
  disdat.row[0].disfield=(double*) malloc(disdat.disrows*disdat.discolumns*sizeof(double));
  if(disdat.row[0].disfield==NULL)
    {
      fprintf(stderr, "not enough memory to copy DISDAT\n");
      exit(1);
    }   
  for(disrow_nr=1;disrow_nr<disdat.disrows;disrow_nr++)
    disdat.row[disrow_nr].disfield=&(disdat.row[0].disfield[disrow_nr*disdat.discolumns]);

  for(disrow_nr=0;disrow_nr<disdat.counter;disrow_nr++)
    for(disfield_nr=0;disfield_nr<disdat.discolumns;disfield_nr++)
      disdat.row[disrow_nr].disfield[disfield_nr]=
	disdat_src.row[disrow_nr].disfield[disfield_nr];

  disdat.class_field=(int*) malloc(disdat.class_field_length*sizeof(int));
  if(disdat.class_field==NULL)
    {
      fprintf(stderr, "not enough memory to copy DISDAT\n");
      exit(1);
    }
 
  for(class_nr=0;class_nr<disdat.class_field_length;class_nr++)
    disdat.class_field[class_nr]=disdat_src.class_field[class_nr];

  return(disdat);
}
