/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "IMG.h"
#include "W2D.h"

double
W2Deuclid_distance (double *f1, double *f2, int dim)
{
  int i;
  double dis = 0.0;

  for (i = 0; i < dim; i++)
    dis += (f1[i] - f2[i]) * (f1[i] - f2[i]);

  return (dis);
}

inline double
pixel_distance (int A_x, int A_y, int B_x, int B_y, int diff_left,
		int diff_under, IMG A, IMG B)
{
  static const double my_sqrt[] = { 0.0, 1.0, 1.4142136 };
  static const double my_pen[] = { 0.0, 1.0, 2.0 };
  static const double binomial3x3[] = { 1, 2, 1,
					2, 4, 2,
					1, 2, 1 };
  static const double binomial5x5[] = {  1,  4,  6,  4,  1,
					 4, 16, 24, 16,  4,
					 6, 24, 36, 24,  6,
					 4, 16, 24, 16,  4,
					 1,  4,  6,  4,  1 };
  static double *A_data;
  static double *B_data;
  static double temp_distance;
  static double distance;
  static double temp_a;
  static double temp_b;
  static int img_depth;
  static int x,y;
  static int centerA,centerB,A_x_size,B_x_size,A_y_size,B_y_size;
  static double tmp0,tmp1,tmp2,tmp3,tmp4,tmp5,tmp6,tmp7,tmp8;

  distance = 0.0;
  img_depth = 0;
  A_x_size=A.x_size;  A_y_size=A.y_size;
  B_x_size=B.x_size;  B_y_size=B.y_size;

  if (B.warprange != -1
      && (B.warprange < 
          abs((int)((double)((B_x_size-1)*A_x)/(double)(A_x_size-1)+0.5)
	      - B_x)
          || B.warprange < 
          abs ((int)((double)((B.y_size-1)*A_y)/(double)(A.y_size-1)+0.5)
               - B_y)))
    return (DBL_MAX);

  while (B.dis_mode[img_depth] != -2) // iteration over image layers; layer after last layer is sentinel and has value -2
    {
      A_data = A.data_list[img_depth];
      B_data = B.data_list[img_depth];
      temp_distance = 0;

      switch (B.dis_mode[img_depth])
	{
	case W2D_euc2:
	  temp_distance =
	    (A_data[(A_y * A_x_size) + A_x] -
	     B_data[(B_y * B_x_size) + B_x]);
	  temp_distance *= temp_distance * B.dis_scale[img_depth];
	  break;
	case W2D_euc2rec3x3:
	  centerA = A_y * A_x_size + A_x;
	  centerB = B_y * B_x_size + B_x;
	  if (     (A_x + 1)<A_x_size && (A_x - 1)>=0 &&  (A_y + 1)<A_y_size && (A_y - 1)>=0 
	       &&  (B_x + 1)<B_x_size && (B_x - 1)>=0 &&  (B_y + 1)<B_y_size && (B_y - 1)>=0  )
	    // everything inside, should happen most often, no more checks necessary
	    {
	      // use possible multiple FPU-pipelines (e.g. on AMD Athlon)
	      tmp0 = A_data[centerA - A_x_size - 1] - B_data[centerB - B_x_size - 1];
	      tmp1 = A_data[centerA - A_x_size    ] - B_data[centerB - B_x_size    ];
	      tmp2 = A_data[centerA - A_x_size + 1] - B_data[centerB - B_x_size + 1];
	      tmp3 = A_data[centerA            - 1] - B_data[centerB            - 1];
	      tmp4 = A_data[centerA               ] - B_data[centerB               ];
	      tmp5 = A_data[centerA            + 1] - B_data[centerB            + 1];
	      tmp6 = A_data[centerA + A_x_size - 1] - B_data[centerB + B_x_size - 1];
	      tmp7 = A_data[centerA + A_x_size    ] - B_data[centerB + B_x_size    ];
	      tmp8 = A_data[centerA + A_x_size + 1] - B_data[centerB + B_x_size + 1];
	      tmp0 *= tmp0;
	      tmp1 *= tmp1;
	      tmp2 *= tmp2;
	      tmp3 *= tmp3;
	      tmp4 *= tmp4;
	      tmp5 *= tmp5;
	      tmp6 *= tmp6;
	      tmp7 *= tmp7;
	      tmp8 *= tmp8;
	      temp_distance = (tmp0 + tmp1 + tmp2 + tmp3 + tmp4 + tmp5 + tmp6 + tmp7 + tmp8) 
		* 0.1111111111111111 * B.dis_scale[img_depth]; 
	    }
	  else // check every pixel
	    {
	      temp_b=0.0;
	      for(x=-1;x<=1;++x)
		{
		  if((A_x + x)>=0 && (A_x + x)<A_x_size && (B_x + x)>=0 && (B_x + x)<B_x_size )
		    {
		      for(y=-1;y<=1;++y)
			{
			  if( (A_y + y)>=0 &&  (A_y + y)<A_y_size  &&  (B_y + y)>=0 &&  (B_y + y)<B_y_size)
			    {
			      temp_a =
				A_data[((A_y + y) * A_x_size) + (A_x + x)] -
				B_data[((B_y + y) * B_x_size) + (B_x + x)];
			      temp_distance+=temp_a*temp_a;
			      temp_b+=1.0;
			    }
			}
		    }
		}
	      temp_distance *= B.dis_scale[img_depth]/temp_b;
	    }
	  break;
	case W2D_euc3rec3x3:
	  temp_b=0.0;
	  for(x=-1;x<=1;x++)
	    for(y=-1;y<=1;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];

		    if(temp_a<0.0)
		      temp_a*=-1.0;

		    temp_distance+=temp_a*temp_a*temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2thresholdrec3x3:
	  temp_b=0.0;
	  for(x=-1;x<=1;x++)
	    for(y=-1;y<=1;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      (A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		       B_data[((B_y + y) * B_x_size) + (B_x + x)])
		      *(A_data[((A_y + y) * A_x_size) + (A_x + x)] -
			B_data[((B_y + y) * B_x_size) + (B_x + x)]);
		    
		    if(temp_a>B.dis_scale[img_depth])
		      temp_a=B.dis_scale[img_depth];
		    
		    temp_distance+=temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance /= temp_b;
	  break;
	case W2D_absrec3x3:
	  temp_b=0.0;
	  for(x=-1;x<=1;x++)
	    for(y=-1;y<=1;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    if(temp_a<0.0)
		      temp_distance+=temp_a*(-1.0);
		    else
		      temp_distance+=temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2rec5x5:
	  temp_b=0.0;
	  for(x=-2;x<=2;x++)
	    for(y=-2;y<=2;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2bin3x3:
	  temp_b=0.0;
	  for(x=-1;x<=1;x++)
	    for(y=-1;y<=1;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a*binomial3x3[((y+1)*3)+x+1];
		    temp_b+=binomial3x3[((y+1)*3)+x+1];
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2bin5x5:
	  temp_b=0.0;
	  for(x=-2;x<=2;x++)
	    for(y=-2;y<=2;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a*binomial5x5[((y+2)*5)+x+2];
		    temp_b+=binomial5x5[((y+2)*5)+x+2];
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_abs:
	  temp_distance = (A_data[(A_y * A_x_size) + A_x] -
			   B_data[(B_y * B_x_size) + B_x]);
	  if(temp_distance<0.0)
	    temp_distance*=-1.0;
	  break;
	case W2D_euc3:
	  temp_distance = 
	      (A_data[(A_y * A_x_size) + A_x] -
	       B_data[(B_y * B_x_size) + B_x])
	    * (A_data[(A_y * A_x_size) + A_x] -
	       B_data[(B_y * B_x_size) + B_x])
	    * (A_data[(A_y * A_x_size) + A_x] -
	       B_data[(B_y * B_x_size) + B_x]);
	  if(temp_distance<0.0)
	    temp_distance*=-1.0;
	  temp_distance*=B.dis_scale[img_depth];
	  break;
	case W2D_pen:
	  temp_distance =
	    (my_sqrt[diff_left] + my_sqrt[diff_under]) * B.dis_scale[img_depth];
	  break;
	case W2D_pen2:
	  temp_distance =
	    (my_pen[diff_left] + my_pen[diff_under]) * B.dis_scale[img_depth];
	  break;
	case W2D_len:
	  temp_distance =
	    sqrt(((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
	    * B.dis_scale[img_depth];
	  break;
	case W2D_len2:
	  temp_distance =
	    (((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
	    * B.dis_scale[img_depth];
	  break;
	case W2D_penborder:
	  if(A_x>1 && A_y>1 && A_x<(A_x_size-2) && A_y<(A_y_size-2))
	    temp_distance =
	      (my_sqrt[diff_left] + my_sqrt[diff_under]) * B.dis_scale[img_depth];
	  break;
	case W2D_pixelthresholdmin:
	  if(distance<=B.dis_scale[img_depth])
	    distance=0;
	  break;
	case W2D_pixelthreshold:
	  if(distance>B.dis_scale[img_depth])
	    distance=B.dis_scale[img_depth];
	  break;
	case W2D_pixeldismove:
	  distance-=B.dis_scale[img_depth];
	  if(distance<0.0)
	    distance*=-1.0;
	  break;
	case W2D_penwarp:
	      if(B_data[(B_y * B_x_size) + B_x] > 0)
		temp_distance=
		  (((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
		  /B_data[(B_y * B_x_size) + B_x];
	      else
		temp_distance=
		  (((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
		  /0.01;	    
	      
	      temp_distance*=B.dis_scale[img_depth];
	  break;
	  /*	case W2D_warpsum:
		if(B_data[(B_y * B_x_size) + B_x] >= 0)
		{
		if(B_data[(B_y * B_x_size) + B_x] > 0)
		temp_distance=
		(((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
		/B_data[(B_y * B_x_size) + B_x];
		else
		temp_distance=
		(((A_x - B_x) * (A_x - B_x)) + ((A_y - B_y) * (A_y - B_y)))
		/0.01;	    
		
		temp_distance*=B.dis_scale[img_depth];
		}
		break;
	  */
	case W2D_penvar:
	  if(B_data[(B_y * B_x_size) + B_x] >= 0)
	    {
	      if(B_data[(B_y * B_x_size) + B_x] > 0)
		temp_distance=distance/B_data[(B_y * B_x_size) + B_x]*B.dis_scale[img_depth];
	      else
		temp_distance=distance/5.0*B.dis_scale[img_depth];
	      if(temp_distance>distance)
		temp_distance=distance*(-1.0);
	      else
		temp_distance*=(-1.0);
	    }
	  break;
	case W2D_euc2rec3x3var:
	  if(B.dis_mode[img_depth+1]==W2D_euc2rec3x3)
	    {
	      temp_b=0.0;
	      for(x=-1;x<=1;x++)
		for(y=-1;y<=1;y++)
		  {
		    if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		       && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		      {
			temp_a =
			  A.data_list[img_depth+1][((A_y + y) * A_x_size) + (A_x + x)] -
			  B.data_list[img_depth+1][((B_y + y) * B_x_size) + (B_x + x)];
			temp_distance+=temp_a*temp_a*B_data[((y+1)*3)+x+1];
			temp_b+=B_data[((y+1)*3)+x+1];
		      }
		  }
	      temp_distance *= B.dis_scale[img_depth]/temp_b;
	    }
	  break;
	case W2D_euc2rec7x7:
	  temp_b=0.0;
	  for(x=-3;x<=3;x++)
	    for(y=-3;y<=3;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2rec9x9:
	  temp_b=0.0;
	  for(x=-4;x<=4;x++)
	    for(y=-4;y<=4;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	case W2D_euc2rec11x11:
	  temp_b=0.0;
	  for(x=-5;x<=5;x++)
	    for(y=-5;y<=5;y++)
	      {
		if((A_x + x)>=0 && (A_y + y)>=0 && (A_x + x)<A_x_size && (A_y + y)<A_y_size 
		   && (B_x + x)>=0 && (B_y + y)>=0 && (B_x + x)<B_x_size && (B_y + y)<B_y_size)
		  {
		    temp_a =
		      A_data[((A_y + y) * A_x_size) + (A_x + x)] -
		      B_data[((B_y + y) * B_x_size) + (B_x + x)];
		    temp_distance+=temp_a*temp_a;
		    temp_b+=1;
		  }
	      }
	  temp_distance *= B.dis_scale[img_depth]/temp_b;
	  break;
	}  // end of switch
      
      distance += temp_distance;
      img_depth++;
    }
  
  return (distance);
}


double W2D_NOTHING(IMG A,IMG B,int* warp_data)
{
  int A_x;
  int A_y;
  int B_x;
  int B_y;

  if(warp_data!=NULL)
    {  
      for(A_y=0;A_y<A.y_size;A_y++)
	for(A_x=0;A_x<A.x_size;A_x++)
	  {
	    B_x=(int) ((double)((B.x_size-1)*A_x)/(double)(A.x_size-1)+0.5);
	    B_y=(int) ((double)((B.y_size-1)*A_y)/(double)(A.y_size-1)+0.5);
	    warp_data[((A.x_size*A_y)+A_x)*2]=B_x;
	    warp_data[(((A.x_size*A_y)+A_x)*2)+1]=B_y;	
	  }
    }
  return(0.0);
}


double
W2DEUC_distance (IMG A, IMG B, int* warp_data)
{
  double dis = 0;
  int A_x, A_y;
  int B_x, B_y;
  int warprange;

  warprange=B.warprange; //needed to compare images different sizes
  B.warprange=-1; //maybe change pixeldistance 
  
  if(A.x_size==B.x_size && A.y_size==B.y_size)
    for (A_y = 0; A_y < A.y_size; A_y++)
      for (A_x = 0; A_x < A.x_size; A_x++)
	{
	  dis += pixel_distance (A_x, A_y, A_x, A_y, 0, 0, A, B);
	  
	  if(warp_data!=NULL)
	    {
	      warp_data[((A_y * A.x_size) + A_x) * 2] = A_x;
	      warp_data[(((A_y * A.x_size) + A_x) * 2) + 1] = A_y;
	    }	  
	}
  else
    for (A_y = 0; A_y < A.y_size; A_y++)
      for (A_x = 0; A_x < A.x_size; A_x++)
	{
	  B_x = (int)((double)((B.x_size-1)*A_x)/(double)(A.x_size-1)+0.5);
	  B_y = (int)((double)((B.y_size-1)*A_y)/(double)(A.y_size-1)+0.5);
	  
	  dis += pixel_distance (A_x, A_y, B_x, B_y, 0, 0, A, B);
	  
	  if(warp_data!=NULL)
	    {
	      warp_data[((A_y * A.x_size) + A_x) * 2] = B_x;
	      warp_data[(((A_y * A.x_size) + A_x) * 2) + 1] = B_y;
	    }
	}

  B.warprange=warprange;  
  return (dis);
}

double
W2DIDM_distance(IMG A, IMG B, int *warp_data)
{
  double dis = 0;
  double best_dis;
  double temp;
  int x,y,xx,yy,best_x=0,best_y=0,B_x,B_y;
  int warprange;

  warprange=B.warprange;
  B.warprange=-1;

  if(A.x_size==B.x_size && A.y_size==B.y_size) 
    for (y = 0; y < A.y_size; y++)
      for (x = 0; x < A.x_size; x++)
	{
	  best_dis=DBL_MAX;
	  for(xx=x-warprange;xx<=x+warprange;xx++)
	    for(yy=y-warprange;yy<=y+warprange;yy++)
	      if(xx >= 0 && yy >= 0 && xx < B.x_size && yy < B.y_size
		 && ((temp=pixel_distance (x, y, xx, yy, 0, 0, A, B))<best_dis
		     || (temp==best_dis && xx==x && yy==y)))
		{
		  best_dis = temp;
		  best_x = xx;
		  best_y = yy;
		}
	  dis += best_dis;
	  if(warp_data!=NULL)
	    {
	      warp_data[((y * A.x_size) + x) * 2] = best_x;
	      warp_data[(((y * A.x_size) + x) * 2) + 1] = best_y;
	    }
	}
  else
    for (y = 0; y < A.y_size; y++)
      for (x = 0; x < A.x_size; x++)
	{
	  best_dis=DBL_MAX;

	  B_x=(int) ((double)((B.x_size-1)*x)/(double)(A.x_size-1)+0.5);
	  B_y=(int) ((double)((B.y_size-1)*y)/(double)(A.y_size-1)+0.5);

	  for(xx=B_x-warprange;xx<=B_x+warprange;xx++)
	    for(yy=B_y-warprange;yy<=B_y+warprange;yy++)
	      {
		if(xx >= 0 && yy >= 0 && xx < B.x_size && yy < B.y_size
		   && ((temp=pixel_distance (x, y, xx, yy, 0, 0, A, B))<best_dis
		       || (temp==best_dis && xx==B_x && yy==B_y)))
		  {
		    best_dis = temp;
		    best_x = xx;
		    best_y = yy;
		  }
	      }
	  dis += best_dis;
	  if(warp_data!=NULL)
	    {
	      warp_data[((y * A.x_size) + x) * 2] = best_x;
	      warp_data[(((y * A.x_size) + x) * 2) + 1] = best_y;
	    }
	}   

  B.warprange=warprange;
  return (dis);
}

double
W2DIDM3_distance(IMG A, IMG B, int *warp_data)
{
  double dis = 0;
  double best_dis;
  double temp;
  int x,y,xx,yy,best_x=0,best_y=0,B_x,B_y;
  int warprange;

  warprange=B.warprange;
  B.warprange=-1;

  if(A.x_size==B.x_size && A.y_size==B.y_size) 
    for (y = 0; y < A.y_size; y++)
      for (x = 0; x < A.x_size; x++)
	{
	  if((x+3)%3==1 && (y+3)%3==1)
	    {
	      best_dis=DBL_MAX;
	      for(xx=x-warprange;xx<=x+warprange;xx++)
		for(yy=y-warprange;yy<=y+warprange;yy++)
		  if(xx >= 0 && yy >= 0 && xx < B.x_size && yy < B.y_size
		     && ((temp=pixel_distance (x, y, xx, yy, 0, 0, A, B))<best_dis
			 || (temp==best_dis && xx==x && yy==y)))
		    {
		      best_dis = temp;
		      best_x = xx;
		      best_y = yy;
		    }
	      dis += best_dis;
	    }
	  else
	    {
	      best_x=x;
	      best_y=y;
	    }
	  if(warp_data!=NULL)
	    {
	      warp_data[((y * A.x_size) + x) * 2] = best_x;
	      warp_data[(((y * A.x_size) + x) * 2) + 1] = best_y;
	    }
	}
  else
    for (y = 0; y < A.y_size; y++)
      for (x = 0; x < A.x_size; x++)
	{

	  B_x=(int) ((double)((B.x_size-1)*x)/(double)(A.x_size-1)+0.5);
	  B_y=(int) ((double)((B.y_size-1)*y)/(double)(A.y_size-1)+0.5);

	  if((x+3)%3==1 && (y+3)%3==1)
	    {
	      best_dis=DBL_MAX;
	      for(xx=B_x-warprange;xx<=B_x+warprange;xx++)
		for(yy=B_y-warprange;yy<=B_y+warprange;yy++)
		  {
		    if(xx >= 0 && yy >= 0 && xx < B.x_size && yy < B.y_size
		       && ((temp=pixel_distance (x, y, xx, yy, 0, 0, A, B))<best_dis
			   || (temp==best_dis && xx==B_x && yy==B_y)))
		      {
			best_dis = temp;
			best_x = xx;
			best_y = yy;
		      }
		  }
	      dis += best_dis;
	    }
	  else
	    {
	      best_x=B_x;
	      best_y=B_y;
	    }
	  if(warp_data!=NULL)
	    {
	      warp_data[((y * A.x_size) + x) * 2] = best_x;
	      warp_data[(((y * A.x_size) + x) * 2) + 1] = best_y;
	    }
	}   
  
  B.warprange=warprange;
  return (dis);
}

double
W2D_img_dis_without_penalty_from_warp (IMG A, IMG B, int *warp_data)
{
  double dis;
  int img_depth = 0;
  IMG B_temp;


  IMG_copy (B, &B_temp);
 
  B_temp.warprange=-1;

  while (B_temp.dis_mode[img_depth] != -2)
    {
      if (B_temp.dis_mode[img_depth] < 1000
	  || B_temp.dis_mode[img_depth] > 2000)
	B_temp.dis_mode[img_depth] = W2D_null;
      //  printf("dismode:%d disscale:%f\n",B_temp.dis_mode[img_depth],B_temp.dis_scale[img_depth]);
      img_depth++;
    }

  dis = W2D_img_dis_from_warp (A, B_temp, warp_data);

  IMG_free (&B_temp);

  return (dis);
}


double
W2D_img_dis_from_warp (IMG A, IMG B, int *warp_data)
{
  int x, y, warp_x, warp_y, left_x, left_y, under_x, under_y;
  double distance;

  distance = 0;
  for (y = 0; y < A.y_size; y++)
    for (x = 0; x < A.x_size; x++)
      {
	warp_x = warp_data[((y * A.x_size) + x) * 2];
	warp_y = warp_data[(((y * A.x_size) + x) * 2) + 1];
	if (x != 0)
	  {
	    left_x = warp_data[((y * A.x_size) + x - 1) * 2];
	    left_y = warp_data[(((y * A.x_size) + x - 1) * 2) + 1];
	  }
	else
	  {
	    left_x = warp_x - 1;
	    left_y = warp_y;
	  }

	if (y != 0)
	  {
	    under_x = warp_data[(((y - 1) * A.x_size) + x) * 2];
	    under_y = warp_data[((((y - 1) * A.x_size) + x) * 2) + 1];
	  }
	else
	  {
	    under_x = warp_x;
	    under_y = warp_y - 1;
	  }
	distance += pixel_distance (x, y, warp_x, warp_y,
				    abs (left_x + 1 - warp_x) + abs (left_y -
								     warp_y),
				    abs (under_x - warp_x) + abs (under_y +
								  1 - warp_y),
				    A, B);
      }
  return (distance);
}

int
W2D_print_warp_data (int *warp_data,int x_size, int y_size)
{
  int x, y, warp_x, warp_y;

  for (y = 0; y < y_size; y++)
    {
      for (x = 0; x < x_size; x++)
	{
	  warp_x = warp_data[((y * x_size) + x) * 2];
	  warp_y = warp_data[(((y * x_size) + x) * 2) + 1];
	  fprintf(stdout,"%3d,%3d ",warp_x,warp_y);
	}
      fprintf(stdout,"\n");
    }
  return (0);
}


int
W2D_calculate_penvar(IMG_list* img_list_tr,IMG_list* img_list_te,double W2D_function(IMG,IMG,int*),int mode)
{
  int var_img_nr;
  int img_nr;
  int* warp_data;
  int* warp_counter;
  double* var_data;
  double* mean_data;
  int A_x;
  int A_y;
  int A_x_size;
  int A_y_size;
  int B_x;
  int B_y;
  int B_x_size;
  int B_y_size; 
  double pix_dis;
  IMG B;
  int img_depth;
  int data_x;

  warp_data=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_te)*2);
  warp_counter=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_tr));  
  var_data=malloc (sizeof (double) * IMG_get_max_dim_list(img_list_tr));
  mean_data=malloc (sizeof (double) * IMG_get_max_dim_list(img_list_tr));
  
  if (warp_data == NULL || warp_counter == NULL
      || var_data == NULL || mean_data == NULL)
    {
      fprintf (stderr, "not enough memory to create penvar data for IMG_list %s\n",
	       img_list_tr->name);
      exit (-1);
    }
  for(var_img_nr=0;var_img_nr<img_list_tr->total_nr;var_img_nr++)
    {
      for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
	    *img_list_tr->img[var_img_nr].y_size;data_x++)
	{
	  var_data[data_x]=0;
	  mean_data[data_x]=0;
	  warp_counter[data_x]=0;
	}
      
      for(img_nr=0;img_nr<img_list_te->total_nr;img_nr++)
	if(img_list_tr->img[var_img_nr].class==img_list_te->img[img_nr].class 
	   && (var_img_nr!=img_nr || img_list_te!=img_list_tr))
	  {
	    IMG_copy (img_list_tr->img[var_img_nr], &B);
	    for(img_depth=0;B.dis_mode[img_depth]!=-2;img_depth++)
	      if (B.dis_mode[img_depth]==W2D_penvar
		  && mode==0)
		B.dis_mode[img_depth] = W2D_null;	

	    W2D_function(img_list_te->img[img_nr],B,warp_data);

	    for(img_depth=0;B.dis_mode[img_depth]!=-2;img_depth++)
	      if (B.dis_mode[img_depth] < 1000
		  || B.dis_mode[img_depth] > 2000)
		B.dis_mode[img_depth] = W2D_null;
	    
	    A_x_size=img_list_te->img[img_nr].x_size;
	    A_y_size=img_list_te->img[img_nr].y_size;
	    B_x_size=img_list_tr->img[var_img_nr].x_size;
	    B_y_size=img_list_tr->img[var_img_nr].y_size;	    
	    for(A_y=0;A_y<A_y_size;A_y++)
	      for(A_x=0;A_x<A_x_size;A_x++)
		{
		  B_x=warp_data[((A_y*A_x_size)+A_x)*2];
		  B_y=warp_data[(((A_y*A_x_size)+A_x)*2)+1];
		  pix_dis=pixel_distance (A_x, A_y, B_x, B_y,			  
					  0, 0, img_list_te->img[img_nr],B);
		  warp_counter[(B_y*B_x_size)+B_x]++;
		  mean_data[(B_y*B_x_size)+B_x]+=pix_dis;
		  var_data[(B_y*B_x_size)+B_x]+=pix_dis*pix_dis;
		}
	    IMG_free(&B);
	  }
      for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
	    *img_list_tr->img[var_img_nr].y_size;data_x++)
	if(warp_counter[data_x]!=0)
	  {
	    mean_data[data_x]/=warp_counter[data_x];
	    var_data[data_x]=(var_data[data_x]/warp_counter[data_x])
	      -(mean_data[data_x]*mean_data[data_x]);
	  }
	else
	  {
	    mean_data[data_x]=-1;
	    var_data[data_x]=-1;
	  }
      for(img_depth=0;img_list_tr->img[var_img_nr].dis_mode[img_depth]!=-2;img_depth++)
	if(img_list_tr->img[var_img_nr].dis_mode[img_depth]==W2D_penvar)
	  for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
		*img_list_tr->img[var_img_nr].y_size;data_x++)
	    img_list_tr->img[var_img_nr].data_list[img_depth][data_x]=
	      mean_data[data_x];
	    
    }
  return(0);
}

int
W2D_calculate_warpsum(IMG_list* img_list_tr,IMG_list* img_list_te,double W2D_function(IMG,IMG,int*),int mode)
{
  int var_img_nr;
  int img_nr;
  int* warp_data;
  int* warp_counter;
  double* var_data;
  double* mean_data;
  int A_x;
  int A_y;
  int A_x_size;
  int A_y_size;
  int B_x;
  int B_y;
  int B_x_size;
  int B_y_size;  
  double square_dis;
  int img_depth;
  int data_x;

  warp_data=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_te)*2);
  warp_counter=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_tr));  
  var_data=malloc (sizeof (double) * IMG_get_max_dim_list(img_list_tr));
  mean_data=malloc (sizeof (double) * IMG_get_max_dim_list(img_list_tr));
  
  if (warp_data == NULL || warp_counter == NULL
      || var_data == NULL || mean_data == NULL)
    {
      fprintf (stderr, "not enough memory to create penvar data for IMG_list %s\n",
	       img_list_tr->name);
      exit (-1);
    }
  for(var_img_nr=0;var_img_nr<img_list_tr->total_nr;var_img_nr++)
    {
      for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
	    *img_list_tr->img[var_img_nr].y_size;data_x++)
	{
	  var_data[data_x]=0;
	  mean_data[data_x]=0;
	  warp_counter[data_x]=0;
	}
      
      for(img_nr=0;img_nr<img_list_te->total_nr;img_nr++)
	if(img_list_tr->img[var_img_nr].class==img_list_te->img[img_nr].class 
	   && (var_img_nr!=img_nr || img_list_te!=img_list_tr))
	  {
	    W2D_function(img_list_te->img[img_nr],img_list_tr->img[var_img_nr],warp_data);
	    
	    A_x_size=img_list_te->img[img_nr].x_size;
	    A_y_size=img_list_te->img[img_nr].y_size;
	    B_x_size=img_list_tr->img[var_img_nr].x_size;
	    B_y_size=img_list_tr->img[var_img_nr].y_size;	    
	    for(A_y=0;A_y<A_y_size;A_y++)
	      for(A_x=0;A_x<A_x_size;A_x++)
		{
		  B_x=warp_data[((A_y*A_x_size)+A_x)*2];
		  B_y=warp_data[(((A_y*A_x_size)+A_x)*2)+1];
		  square_dis=
		    ( (B_x-A_x) * (B_x-A_x) )
		    + ( (B_y-A_y) * (B_y-A_y) );
		  warp_counter[(B_y*B_x_size)+B_x]++;
		  mean_data[(B_y*B_x_size)+B_x]+=square_dis;
		  var_data[(B_y*B_x_size)+B_x]+=square_dis*square_dis;
		}
	  }
      for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
	    *img_list_tr->img[var_img_nr].y_size;data_x++)
	if(warp_counter[data_x]!=0)
	  {
	    mean_data[data_x]/=warp_counter[data_x];
	    var_data[data_x]=(var_data[data_x]/warp_counter[data_x])
	      -(mean_data[data_x]*mean_data[data_x]);
	  }
	else
	  {
	    mean_data[data_x]=0;
	    var_data[data_x]=0;
	  }
      for(img_depth=0;img_list_tr->img[var_img_nr].dis_mode[img_depth]!=-2;img_depth++)
	if(img_list_tr->img[var_img_nr].dis_mode[img_depth]==W2D_warpsum)
	  for(data_x=0;data_x<img_list_tr->img[var_img_nr].x_size
		*img_list_tr->img[var_img_nr].y_size;data_x++)
	    img_list_tr->img[var_img_nr].data_list[img_depth][data_x]=
	      mean_data[data_x];
    }
  free(var_data);
  free(mean_data);
  free(warp_counter);
  free(warp_data);
  return(0);
}

int
W2D_calculate_dis5x5_var(IMG_list* img_list_tr, IMG_list* img_list_te, double W2D_function(IMG,IMG,int*))
{
  int img_nr;
  int tr_img_nr;
  int* warp_data;
  long long unsigned int* dis5x5_counter;
  double* var_data;
  double* mean_data;
  IMG B;
  double double_temp;
  int A_x, A_y, B_x, B_y, x, y;
  int A_x_size, A_y_size, B_x_size, B_y_size;
  int img_depth;

  warp_data=malloc (sizeof (int) * IMG_get_max_dim_list(img_list_te)*2);
  dis5x5_counter=malloc (sizeof (long long unsigned int) * 25);  
  var_data=malloc (sizeof (double) * 25);
  mean_data=malloc (sizeof (double) * 25);
  
  if (warp_data == NULL || dis5x5_counter == NULL
      || var_data == NULL || mean_data == NULL)
    {
      fprintf (stderr, "not enough memory to create penvar data for IMG_list %s\n",
	       img_list_tr->name);
      exit (-1);
    }

  for(x=0;x<25;x++)
    {
      var_data[x]=0;
      mean_data[x]=0;
      dis5x5_counter[x]=0;
    }

  for(img_nr=0;img_nr<img_list_te->total_nr;img_nr++)
    {
      for(tr_img_nr=0;tr_img_nr<img_list_tr->total_nr;tr_img_nr++)
	{      
	  if((img_nr!=tr_img_nr || img_list_te!=img_list_tr)
	     && img_list_te->img[img_nr].class==img_list_tr->img[tr_img_nr].class)
	    {
	      W2D_function(img_list_te->img[img_nr],img_list_tr->img[tr_img_nr],warp_data);
	      IMG_copy (img_list_tr->img[tr_img_nr], &B);
	      for(img_depth=0;B.dis_mode[img_depth]!=-2;img_depth++)
		if (B.dis_mode[img_depth] < 1000
		    || B.dis_mode[img_depth] > 2000)
		  {
		    if(B.dis_mode[img_depth]==W2D_euc2rec3x3var
		       && B.dis_mode[img_depth+1]==W2D_euc2rec3x3
		       && B.dis_scale[img_depth+1]==0.0)
		      B.dis_scale[img_depth+1]=B.dis_scale[img_depth];
		    B.dis_mode[img_depth] = W2D_null;
		  }
		else
		  {
		    B.dis_mode[img_depth] = W2D_euc2;
		  }
	      A_x_size=img_list_te->img[img_nr].x_size;
	      A_y_size=img_list_te->img[img_nr].y_size;
	      B_x_size=img_list_tr->img[tr_img_nr].x_size;
	      B_y_size=img_list_tr->img[tr_img_nr].y_size;
	      for(A_y=0;A_y<A_y_size;A_y++)
		for(A_x=0;A_x<A_x_size;A_x++)
		  {
		    B_y=warp_data[(((A_x_size*A_y)+A_x)*2)+1];
		    B_x=warp_data[(((A_x_size*A_y)+A_x)*2)];
		    for(y=-2;y<=2;y++)
		      for(x=-2;x<=2;x++)
			{
			  if(B_y+y>=0 && B_y+y<B_y_size &&
			     B_x+x>=0 && B_x+x<B_x_size)
			    {
			      dis5x5_counter[(y*5)+x+12]++;
			      double_temp=pixel_distance (A_x, A_y, B_x, B_y,			  
							  0, 0, img_list_te->img[img_nr],B);
			      mean_data[(y*5)+x+12]+=double_temp;
			      var_data[(y*5)+x+12]+=double_temp*double_temp;
			    }
			}
		  }
	      IMG_free(&B);
	    }
	}
    }


  printf("#dis5x5_counter\n");
  for(y=-2;y<=2;y++)
    {
      printf("#");
      for(x=-2;x<=2;x++)
	{
	  if(dis5x5_counter[(y*5)+x+12]!=0)
	    {
	      mean_data[(y*5)+x+12]/=dis5x5_counter[(y*5)+x+12];
	      var_data[(y*5)+x+12]=(var_data[(y*5)+x+12]/dis5x5_counter[(y*5)+x+12])-
		(mean_data[(y*5)+x+12]*mean_data[(y*5)+x+12]);
	    }
	  else
	    {
	      mean_data[(y*5)+x+12]=-1;
	      var_data[(y*5)+x+12]=-1;
	    }
	  printf("%d ",dis5x5_counter[(y*5)+x+12]);
	}
      printf("\n");
    }
  
  printf("#mean_data\n");
  for(y=-2;y<=2;y++)
    {
      printf("#");
      for(x=-2;x<=2;x++)
	printf("%f ",mean_data[(y*5)+x+12]);
      printf("\n");
    }

  printf("#var_data\n");
  for(y=-2;y<=2;y++)
    {
      printf("#");
      for(x=-2;x<=2;x++)
	printf("%f ",var_data[(y*5)+x+12]);
      printf("\n");
    }


  double_temp=0;
    for(x=0;x<25;x++)
      {
	mean_data[x]=1/mean_data[x];
	if(mean_data[x]>double_temp)
	  double_temp=mean_data[x];
      }
  printf("#scale_data\n");
    for(y=-2;y<=2;y++)
      {
	printf("#");
      for(x=-2;x<=2;x++)
	{
	  mean_data[(y*5)+x+12]/=double_temp;
	  printf("%f ",mean_data[(y*5)+x+12]);  
	}
      printf("\n");
      }
    for(tr_img_nr=0;tr_img_nr<img_list_tr->total_nr;tr_img_nr++)
      {     
	for(img_depth=0;img_list_tr->img[tr_img_nr].dis_mode[img_depth]!=-2;img_depth++)
	  {
	    if (img_list_tr->img[tr_img_nr].dis_mode[img_depth]==W2D_euc2rec3x3var)
	      {
		for(y=-1;y<=1;y++)
		  for(x=-1;x<=1;x++)
		    img_list_tr->img[tr_img_nr].data_list[img_depth][(y*3)+x+4]=mean_data[(y*5)+x+12];
		
	      }
	  }
      }
  
    free(warp_data);
    free(mean_data);
    free(var_data); 
    free(dis5x5_counter);
    return(0);
}
