/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<float.h>
#include"IMG.h"
#include"W2D.h"

double
W2DHM_best_line (IMG A, IMG B, int A_x, int B_x, int x_move, double** dis_field, int** move_field,
		 int *warping_info)
{
  double distance;
  int A_y;
  int B_y;
  int A_y_end;
  int A_y_start;
  double temp;
  double temp_penalty;

  A_y_start=-1;
  A_y_end=A.y_size-1-((B.y_size+1)/2);
  for(A_y=0;A_y<=A_y_end;A_y++) 
    {
	dis_field[-2][A_y]=DBL_MAX;
	dis_field[-1][A_y]=DBL_MAX;
    }
  dis_field[0][0]=pixel_distance (A_x, 0, B_x, 0, x_move, 0, A, B);
  for(B_y=0;B_y<B.y_size;B_y++) 
    {
	if(B_y>1)
  		A_y_start+=B_y%2;
	A_y_end+=(B.y_size-B_y)%2;
	if(A_y_start>=0)
	   dis_field[B_y][A_y_start]=DBL_MAX;
	else
		A_y_start++;
	for(A_y=A_y_start;A_y<A_y_end;A_y++)
	{
		temp =
	    	pixel_distance (A_x, A_y+1, B_x, B_y, x_move, 0,A,B);
		temp_penalty =
	    	pixel_distance (A_x, A_y+1, B_x, B_y, x_move, 1,A,B);

	  if (dis_field[B_y-2][A_y] < dis_field[B_y][A_y])
	    {
	      if (dis_field[B_y-2][A_y]+temp_penalty < dis_field[B_y-1][A_y]+temp)
		{
		  move_field[B_y][A_y+1] = 2;
		  dis_field[B_y][A_y+1] = dis_field[B_y-2][A_y]+temp_penalty;
		}
	      else
		{
		  move_field[B_y][A_y+1] = 1;
		  dis_field[B_y][A_y+1] = dis_field[B_y-1][A_y]+temp;
		}
	    }
	  else
	    {
	      if (dis_field[B_y][A_y]+temp_penalty < dis_field[B_y-1][A_y]+temp)
		{
		  move_field[B_y][A_y+1] = 0;
		  dis_field[B_y][A_y+1] = dis_field[B_y][A_y]+temp_penalty;
		}
	      else
		{
		  move_field[B_y][A_y+1] = 1;
		  dis_field[B_y][A_y+1] = dis_field[B_y-1][A_y]+temp;
		}
	    }
	}
    }
   distance=dis_field[B.y_size-1][A.y_size-1];

   if(warping_info!=NULL)
     {
	B_y=B.y_size-1;
	for(A_y=A.y_size-1;A_y>=0;A_y--)
	{
	        warping_info[((A_x+(A_y*A.x_size))*2)]=B_x;
	        warping_info[((A_x+(A_y*A.x_size))*2)+1]=B_y;
		B_y-=move_field[B_y][A_y];	
	}
     }

   return(distance);
}

double
W2DHM_distance (IMG A, IMG B, int *warping_info)
{
  double distance;
  int A_x,A_y;
  int B_x,B_y;
  int A_x_end;
  int A_x_start;
  double temp;
  double temp_penalty;
  double** dis_field;
  int** move_field;
  double** dis_field_y;
  int** move_field_y;

  if(A.x_size*2<=B.x_size || A.y_size*2<=B.y_size)
  {
    distance=0;
    for (A_y = 0; A_y < A.y_size; A_y++)
      for (A_x = 0; A_x < A.x_size; A_x++)
	{
 	  B_x=(int)((double)((B.x_size-1)*A_x)/(double)(A.x_size-1)+0.5);
	  B_y=(int)((double)((B.y_size-1)*A_y)/(double)(A.y_size-1)+0.5);
          distance += pixel_distance (A_x, A_y, B_x, B_y, 0, 0, A, B);
	  if(warping_info!=NULL)
	     {
	  	warping_info[((A_x+(A_y*A.x_size))*2)]=B_x;
	        warping_info[((A_x+(A_y*A.x_size))*2)+1]=B_y;
	     }
	}
    return(distance);
  }

 dis_field_y = malloc (sizeof (double*) * (B.y_size+2));
  if(dis_field_y==NULL)
	exit -1;
  dis_field_y[0] = malloc (sizeof (double) * A.y_size * (B.y_size+2));
  if(dis_field_y[0]==NULL)
	exit -1;
  move_field_y = malloc (sizeof (int*) * B.y_size);
  if(move_field_y==NULL)
	exit -1;
  move_field_y[0] = malloc (sizeof (int) * A.y_size * B.y_size);
  if (move_field_y[0] == NULL)
	exit -1;
  for(B_x=1;B_x<B.y_size;B_x++)
    {
	dis_field_y[B_x]=&dis_field_y[0][B_x*A.y_size];
	move_field_y[B_x]=&move_field_y[0][B_x*A.y_size];
    }
    dis_field_y[B.y_size]=&dis_field_y[0][B.y_size*A.y_size];
    dis_field_y[B.y_size+1]=&dis_field_y[B.y_size][A.y_size];
    dis_field_y=&dis_field_y[2];

  dis_field = malloc (sizeof (double*) * (B.x_size+2));
  if(dis_field==NULL)
	exit -1;
  dis_field[0] = malloc (sizeof (double) * A.x_size * (B.x_size+2));
  if(dis_field[0]==NULL)
	exit -1;
  move_field = malloc (sizeof (int*) * B.x_size);
  if(move_field==NULL)
	exit -1;
  move_field[0] = malloc (sizeof (int) * A.x_size * B.x_size);
  if (move_field[0] == NULL)
	exit -1;
  for(B_x=1;B_x<B.x_size;B_x++)
    {
	dis_field[B_x]=&dis_field[0][B_x*A.x_size];
	move_field[B_x]=&move_field[0][B_x*A.x_size];
    }
    dis_field[B.x_size]=&dis_field[0][B.x_size*A.x_size];
    dis_field[B.x_size+1]=&dis_field[B.x_size][A.x_size];
    dis_field=&dis_field[2];

  A_x_start=-1;
  A_x_end=A.x_size-1-((B.x_size+1)/2);
  for(A_x=0;A_x<=A_x_end;A_x++) 
    {
	dis_field[-2][A_x]=DBL_MAX;
	dis_field[-1][A_x]=DBL_MAX;
    }
  dis_field[0][0]=W2DHM_best_line(A,B,0,0,0,dis_field_y,move_field_y,NULL);
  for(B_x=0;B_x<B.x_size;B_x++) 
    {
	if(B_x>1)
  		A_x_start+=B_x%2;
	A_x_end+=(B.x_size-B_x)%2;
	if(A_x_start>=0)
	   dis_field[B_x][A_x_start]=DBL_MAX;
	else
		A_x_start++;
	for(A_x=A_x_start;A_x<A_x_end;A_x++)
	{
		temp=W2DHM_best_line(A,B,A_x+1,B_x,0,dis_field_y,move_field_y,NULL);
		temp_penalty=W2DHM_best_line(A,B,A_x+1,B_x,1,dis_field_y,move_field_y,NULL);

	  if (dis_field[B_x-2][A_x] < dis_field[B_x][A_x])
	    {
	      if (dis_field[B_x-2][A_x]+temp_penalty < dis_field[B_x-1][A_x]+temp)
		{
		  move_field[B_x][A_x+1] = 2;
		  dis_field[B_x][A_x+1] = dis_field[B_x-2][A_x]+temp_penalty;
		}
	      else
		{
		  move_field[B_x][A_x+1] = 1;
		  dis_field[B_x][A_x+1] = dis_field[B_x-1][A_x]+temp;
		}
	    }
	  else
	    {
	      if (dis_field[B_x][A_x]+temp_penalty < dis_field[B_x-1][A_x]+temp)
		{
		  move_field[B_x][A_x+1] = 0;
		  dis_field[B_x][A_x+1] = dis_field[B_x][A_x]+temp_penalty;
		}
	      else
		{
		  move_field[B_x][A_x+1] = 1;
		  dis_field[B_x][A_x+1] = dis_field[B_x-1][A_x]+temp;
		}
	   }
	}
    }
   distance=dis_field[B.x_size-1][A.x_size-1];

   if(warping_info!=NULL)
     {
	B_x=B.x_size-1;
	for(A_x=A.x_size-1;A_x>=0;A_x--)
	{
		W2DHM_best_line(A,B,A_x,B_x,(move_field[B_x][A_x]+1)%2,dis_field_y,move_field_y,warping_info);
		B_x-=move_field[B_x][A_x];	
	}
     }

   dis_field=&dis_field[-2];
   free(dis_field[0]);
   free(move_field[0]);
   free(dis_field);
   free(move_field);
   dis_field_y=&dis_field_y[-2];
   free(dis_field_y[0]);
   free(move_field_y[0]);
   free(dis_field_y);
   free(move_field_y);

   return(distance);
}

/***********************************************************************************/

double
W2DHMD_best_line (IMG A, IMG B, int A_x, int B_x, int x_move, double** dis_field, int** move_field,
		  int *warping_info)
{
  double distance;
  int A_y;
  int B_y;
  int A_y_end;
  int A_y_start;
  double temp;
  double temp_penalty;
  int distortion;
  int B_x_temp;

  A_y_start=-1;
  A_y_end=A.y_size-1-((B.y_size+1)/2);
  for(A_y=0;A_y<=A_y_end;A_y++) 
    {
	dis_field[-2][A_y]=DBL_MAX;
	dis_field[-1][A_y]=DBL_MAX;
    }
  dis_field[0][0]=pixel_distance (A_x, 0, B_x, 0, x_move, 0, A, B); 
  for(distortion=-1;distortion<=1;distortion+=2)
    if( (distortion + B_x) >= 0 && (distortion + B_x) < B.x_size
	&& dis_field[0][0] > pixel_distance (A_x, 0, distortion + B_x, 0, x_move, 0, A, B))
      dis_field[0][0]=pixel_distance (A_x, 0, distortion + B_x, 0, x_move, 0, A, B);

  for(B_y=0;B_y<B.y_size;B_y++) 
    {
	if(B_y>1)
  		A_y_start+=B_y%2;
	A_y_end+=(B.y_size-B_y)%2;
	if(A_y_start>=0)
	   dis_field[B_y][A_y_start]=DBL_MAX;
	else
		A_y_start++;
	for(A_y=A_y_start;A_y<A_y_end;A_y++)
	{
	  temp =
	    pixel_distance (A_x, A_y+1, B_x, B_y, x_move, 0,A,B);
	  for(distortion=-1;distortion<=1;distortion+=2)
	    if( (distortion + B_x) >= 0 && (distortion + B_x) < B.x_size
		&& temp > pixel_distance (A_x, A_y+1, distortion + B_x, B_y, x_move, 0, A, B))
	      temp = pixel_distance (A_x, A_y+1, distortion + B_x, B_y, x_move, 0, A, B);
	  temp_penalty =
	    pixel_distance (A_x, A_y+1, B_x, B_y, x_move, 1,A,B);
	  for(distortion=-1;distortion<=1;distortion+=2)
	    if( (distortion + B_x) >= 0 && (distortion + B_x) < B.x_size
		&& temp_penalty > pixel_distance (A_x, A_y+1, distortion + B_x, B_y, x_move, 1,A,B))
	      temp_penalty = pixel_distance (A_x, A_y+1, distortion + B_x, B_y, x_move, 1,A,B);

	  if (dis_field[B_y-2][A_y] < dis_field[B_y][A_y])
	    {
	      if (dis_field[B_y-2][A_y]+temp_penalty < dis_field[B_y-1][A_y]+temp)
		{
		  move_field[B_y][A_y+1] = 2;
		  dis_field[B_y][A_y+1] = dis_field[B_y-2][A_y]+temp_penalty;
		}
	      else
		{
		  move_field[B_y][A_y+1] = 1;
		  dis_field[B_y][A_y+1] = dis_field[B_y-1][A_y]+temp;
		}
	    }
	  else
	    {
	      if (dis_field[B_y][A_y]+temp_penalty < dis_field[B_y-1][A_y]+temp)
		{
		  move_field[B_y][A_y+1] = 0;
		  dis_field[B_y][A_y+1] = dis_field[B_y][A_y]+temp_penalty;
		}
	      else
		{
		  move_field[B_y][A_y+1] = 1;
		  dis_field[B_y][A_y+1] = dis_field[B_y-1][A_y]+temp;
		}
	    }
	}
    }
   distance=dis_field[B.y_size-1][A.y_size-1];

   if(warping_info!=NULL)
     {
	B_y=B.y_size-1;
	for(A_y=A.y_size-1;A_y>=0;A_y--)
	{
	  //change y_move to use all penalty_functions
	  B_x_temp=B_x;
	  temp=pixel_distance (A_x, A_y, B_x, B_y, x_move, 0, A, B);
	  for(distortion=-1;distortion<=1;distortion+=2)
	    if(temp > pixel_distance (A_x, A_y, distortion + B_x, B_y, x_move, 0, A, B))
	      {
		B_x_temp=distortion + B_x;
		temp=pixel_distance (A_x, A_y, distortion + B_x, B_y, x_move, 0, A, B);
	      }
	  
	        warping_info[((A_x+(A_y*A.x_size))*2)]=B_x_temp;
	        warping_info[((A_x+(A_y*A.x_size))*2)+1]=B_y;
		B_y-=move_field[B_y][A_y];	
	}
     }

   return(distance);
}


double
W2DHMD_distance (IMG A, IMG B, int *warping_info)
{
  double distance;
  int A_x,A_y;
  int B_x,B_y;
  int A_x_end;
  int A_x_start;
  double temp;
  double temp_penalty;
  double** dis_field;
  int** move_field;
  double** dis_field_y;
  int** move_field_y;

  if(A.x_size*2<=B.x_size || A.y_size*2<=B.y_size)
  {
    distance=0;
    for (A_y = 0; A_y < A.y_size; A_y++)
      for (A_x = 0; A_x < A.x_size; A_x++)
	{
 	  B_x=(int)((double)((B.x_size-1)*A_x)/(double)(A.x_size-1)+0.5);
	  B_y=(int)((double)((B.y_size-1)*A_y)/(double)(A.y_size-1)+0.5);
          distance += pixel_distance (A_x, A_y, B_x, B_y, 0, 0, A, B);
	  if(warping_info!=NULL)
	     {
	  	warping_info[((A_x+(A_y*A.x_size))*2)]=B_x;
	        warping_info[((A_x+(A_y*A.x_size))*2)+1]=B_y;
	     }
	}
    return(distance);
  }

 dis_field_y = malloc (sizeof (double*) * (B.y_size+2));
  if(dis_field_y==NULL)
	exit -1;
  dis_field_y[0] = malloc (sizeof (double) * A.y_size * (B.y_size+2));
  if(dis_field_y[0]==NULL)
	exit -1;
  move_field_y = malloc (sizeof (int*) * B.y_size);
  if(move_field_y==NULL)
	exit -1;
  move_field_y[0] = malloc (sizeof (int) * A.y_size * B.y_size);
  if (move_field_y[0] == NULL)
	exit -1;
  for(B_x=1;B_x<B.y_size;B_x++)
    {
	dis_field_y[B_x]=&dis_field_y[0][B_x*A.y_size];
	move_field_y[B_x]=&move_field_y[0][B_x*A.y_size];
    }
    dis_field_y[B.y_size]=&dis_field_y[0][B.y_size*A.y_size];
    dis_field_y[B.y_size+1]=&dis_field_y[B.y_size][A.y_size];
    dis_field_y=&dis_field_y[2];

  dis_field = malloc (sizeof (double*) * (B.x_size+2));
  if(dis_field==NULL)
	exit -1;
  dis_field[0] = malloc (sizeof (double) * A.x_size * (B.x_size+2));
  if(dis_field[0]==NULL)
	exit -1;
  move_field = malloc (sizeof (int*) * B.x_size);
  if(move_field==NULL)
	exit -1;
  move_field[0] = malloc (sizeof (int) * A.x_size * B.x_size);
  if (move_field[0] == NULL)
	exit -1;
  for(B_x=1;B_x<B.x_size;B_x++)
    {
	dis_field[B_x]=&dis_field[0][B_x*A.x_size];
	move_field[B_x]=&move_field[0][B_x*A.x_size];
    }
    dis_field[B.x_size]=&dis_field[0][B.x_size*A.x_size];
    dis_field[B.x_size+1]=&dis_field[B.x_size][A.x_size];
    dis_field=&dis_field[2];

  A_x_start=-1;
  A_x_end=A.x_size-1-((B.x_size+1)/2);
  for(A_x=0;A_x<=A_x_end;A_x++) 
    {
	dis_field[-2][A_x]=DBL_MAX;
	dis_field[-1][A_x]=DBL_MAX;
    }
  dis_field[0][0]=W2DHMD_best_line(A,B,0,0,0,dis_field_y,move_field_y,NULL);
  for(B_x=0;B_x<B.x_size;B_x++) 
    {
	if(B_x>1)
  		A_x_start+=B_x%2;
	A_x_end+=(B.x_size-B_x)%2;
	if(A_x_start>=0)
	   dis_field[B_x][A_x_start]=DBL_MAX;
	else
		A_x_start++;
	for(A_x=A_x_start;A_x<A_x_end;A_x++)
	{
		temp=W2DHMD_best_line(A,B,A_x+1,B_x,0,dis_field_y,move_field_y,NULL);
		temp_penalty=W2DHMD_best_line(A,B,A_x+1,B_x,1,dis_field_y,move_field_y,NULL);

	  if (dis_field[B_x-2][A_x] < dis_field[B_x][A_x])
	    {
	      if (dis_field[B_x-2][A_x]+temp_penalty < dis_field[B_x-1][A_x]+temp)
		{
		  move_field[B_x][A_x+1] = 2;
		  dis_field[B_x][A_x+1] = dis_field[B_x-2][A_x]+temp_penalty;
		}
	      else
		{
		  move_field[B_x][A_x+1] = 1;
		  dis_field[B_x][A_x+1] = dis_field[B_x-1][A_x]+temp;
		}
	    }
	  else
	    {
	      if (dis_field[B_x][A_x]+temp_penalty < dis_field[B_x-1][A_x]+temp)
		{
		  move_field[B_x][A_x+1] = 0;
		  dis_field[B_x][A_x+1] = dis_field[B_x][A_x]+temp_penalty;
		}
	      else
		{
		  move_field[B_x][A_x+1] = 1;
		  dis_field[B_x][A_x+1] = dis_field[B_x-1][A_x]+temp;
		}
	   }
	}
    }
   distance=dis_field[B.x_size-1][A.x_size-1];

   if(warping_info!=NULL)
     {
	B_x=B.x_size-1;
	for(A_x=A.x_size-1;A_x>=0;A_x--)
	{
		W2DHMD_best_line(A,B,A_x,B_x,(move_field[B_x][A_x]+1)%2,dis_field_y,move_field_y,warping_info);
		B_x-=move_field[B_x][A_x];	
	}
     }

   dis_field=&dis_field[-2];
   free(dis_field[0]);
   free(move_field[0]);
   free(dis_field);
   free(move_field);
   dis_field_y=&dis_field_y[-2];
   free(dis_field_y[0]);
   free(move_field_y[0]);
   free(dis_field_y);
   free(move_field_y);

   return(distance);
}

