/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<time.h>
#include"IMG.h"
#include"W2D.h"

int
check_warp_data (int *warp_data, int x_size, int y_size)
{
  int x, y, x_move, y_move;
  int offset;

  for (x = 0; x < x_size; x++)
    {
      //check upper_bound
      if (warp_data[(x * 2) + 1] != 0)
	return 0;
      //check lower_bound
      if (warp_data[(((x_size * (y_size - 1)) + x) * 2) + 1] != (y_size - 1))
	return 0;
    }
  for (y = 0; y < y_size; y++)
    {
      //check left_bound
      if (warp_data[(x_size * y * 2)] != 0)
	return 0;
      //check right_bound
      if (warp_data[(((x_size * (y + 1)) - 1) * 2)] != (x_size - 1))
	return 0;
    }

  //printf("\n\n");
  offset = 0;
  for (y = 0; y < y_size - 1; y++)
    {
      for (x = 0; x < x_size - 1; x++)
	{
	  //check right neighbour
	  x_move = warp_data[offset + 2] - warp_data[offset];
	  y_move = warp_data[offset + 3] - warp_data[offset+1];
	  if (x_move < 0 || y_move < -1 || x_move > 2 || y_move > 1)
	    return 0;
	  //check lower neighbour
	  x_move =
	    warp_data[offset + (x_size * 2)] - warp_data[offset];
	  y_move =
	    warp_data[offset + (x_size * 2) + 1] - warp_data[offset + 1];
	  if (x_move < -1 || y_move < 0 || x_move > 1 || y_move > 2)
	    return 0;
	  offset += 2;
	  //printf("(%i,%i) ",x_move,y_move);
	}
      offset += 2;
      //printf("\n");
    }

  return 1;
}

int
init_random ()
{

  srandom (time (NULL));

  return (1);
}

int
equal_random (int min, int max)
{
  return (0.5 + min + ((double) random () * (max - min) / (double) RAND_MAX));
}

double
W2DSA_distance (IMG A, IMG B, int *warping_info)
{
  int max_iter;
  int check = 0;
  int iter;
  int x, y;
  double distance;
  double temp_distance;

  int *warp_data;
  int *temp_warp_data;
  int *int_ptr;

  int block_x_size;
  int block_y_size;
  int block_x_start;
  int block_y_start;
  int block_x_move;
  int block_y_move;

  max_iter = B.beamsize;
  //printf("max_iter:%d\n",max_iter);

  if (NULL ==
      (temp_warp_data = malloc (sizeof (int) * A.x_size * A.y_size * 2))
      || NULL == (warp_data =
		  malloc (sizeof (int) * A.x_size * A.y_size * 2)))
    {
      printf ("error: no memory for warp_data or temp_warp_data\n");
      exit (1);
    }

  // init warp_data with id_warping (means: no warping at all)
  for (x = 0; x < A.x_size * A.y_size; x++)
    {
      warp_data[x * 2] = x % A.x_size;
      warp_data[(x * 2) + 1] = x / A.x_size;
    }

  init_random ();

  iter = 0;
  distance = W2D_img_dis_from_warp (A, B, warp_data);
  while (iter < max_iter)
    {
      iter++;

      //make a copy of warp_data in temp_warp_data
      for (x = 0; x < A.x_size * A.y_size * 2; x++)
	temp_warp_data[x] = warp_data[x];

      //get random block
      block_x_size = equal_random (1, A.x_size);
      block_y_size = equal_random (1, A.y_size);
      block_x_start = equal_random (0, A.x_size - block_x_size);
      block_y_start = equal_random (0, A.y_size - block_y_size);

      block_x_move = 0;
      block_y_move = 0;
      while (block_x_move == 0 && block_y_move == 0)
	{
	  block_x_move = equal_random (-2, 1);
	  block_y_move = equal_random (-2, 1);
	}

      // printf("BLOCK x_size:%d y_size:%d x_start:%d y_start:%d x_move:%d y_move:%d\n",block_x_size,block_y_size,block_x_start,block_y_start,block_x_move,block_y_move);

      //manipulate temp_warp_data 
      for (y = block_y_start; y < block_y_size + block_y_start; y++)
	for (x = block_x_start; x < block_x_size + block_x_start; x++)
	  {
	    //printf("(%d,%d)+%d+%d ",x,y,block_x_move,block_y_move);
	    temp_warp_data[((A.x_size * y) + x) * 2] += block_x_move;
	    temp_warp_data[(((A.x_size * y) + x) * 2) + 1] += block_y_move;
	  }

      //check temp_warp_data
      if (check_warp_data (temp_warp_data, A.x_size, A.y_size))
	{
	  check++;
	  temp_distance = W2D_img_dis_from_warp (A, B, temp_warp_data);
	  //      printf("iter:%d temp_distance:%f\n",iter,temp_distance);
	  if (temp_distance < distance)	// || iter<(max_iter/2))
	    {
	      distance = temp_distance;
	      int_ptr = warp_data;
	      warp_data = temp_warp_data;
	      temp_warp_data = int_ptr;
	    }
	}
    }

  if (warping_info != NULL)
    {
      for (x = 0; x < A.x_size * A.y_size * 2; x++)
	warping_info[x] = warp_data[x];
    }
  //printf("check:%d\n",check);
  free (temp_warp_data);
  free (warp_data);

  return (distance);
}
