/******************************************************************************

Copyright 2003, 2004 Christian Gollan

This file is part of the W2D package.

The W2D package is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

The W2D package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with the W2D package; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include "IMG.h"
#include "W2D.h"

static long unsigned int POW9[32];	//in use for faster hash_index

typedef struct _ww_list
{
  struct _ww_list *previous;
  char *data;
  int ww_data_length;
  int counter;
  int x;
  int y;
  int y_size;
  int beamsize;
}
ww_list;

typedef struct _hash_el
{
  int position;
  struct _hash_el *next;
}
hash_el;

typedef struct _hash_list
{
  hash_el **data;
  int length;
}
hash_list;

typedef struct _list_element
{
  double distance;
  int m_pos;
  int n_pos;
  int ww_pos;
  int last_move;
}
list_element;

list_element *
init_list (int counter)
{
  list_element *list;
  list = malloc (sizeof (list_element) * counter);
  list[0].distance = -1;
  return (list);

}

int
free_list (list_element * list)
{
  free (list);
  return -1;
}

int
push_list (list_element * list, list_element element)
{
  int x = 0;
  while (list[x].distance != -1)
    x++;

  list[x] = element;
  list[x + 1].distance = -1;
  return -1;
}

list_element
pop_list (list_element * list)
{
  int x;
  int best_no=0;
  double best_dis = DBL_MAX;
  list_element element;

  x = 0;
  while (list[x].distance != -1)
    {
      if (list[x].distance <= best_dis)
	{
	  best_no = x;
	  best_dis = list[x].distance;
	}
      x++;
    }
  element = list[best_no];
  list[best_no] = list[x - 1];
  list[x - 1].distance = -1;
  return (element);
}

inline int
read_y_ww (int position, char *data)
{
  return (((data[13 + (position / 2)]) & (3 << (4 * ((position + 1) % 2)))) >>
	  (4 * ((position + 1) % 2)));
}

inline int
read_x_ww (int position, char *data)
{
  return (((data[13 + (position / 2)]) & (12 << (4 * ((position + 1) % 2))))
	  >> (2 + (4 * ((position + 1) % 2))));
}

inline
void write_y_ww (int position, char *data, int y)
{
  data[13 + (position / 2)] =
    data[13 + (position / 2)] & (255 - (3 << (4 * ((position + 1) % 2))));
  data[13 + (position / 2)] =
    data[13 + (position / 2)] | (y << (4 * ((position + 1) % 2)));
}

inline
void write_x_ww (int position, char *data, int x)
{
  data[13 + (position / 2)] =
    data[13 + (position / 2)] & (255 - (12 << (4 * ((position + 1) % 2))));
  data[13 + (position / 2)] =
    data[13 + (position / 2)] | (x << (2 + (4 * ((position + 1) % 2))));
}



double
init_ww_list (ww_list ** wake_list, IMG A, IMG B, char *data)
{
  int M;
  int N;
  int m;
  int n;
  int n_end;
  int n_start;
  int ww_pos;
  double *dis;
  double temp;
  double temp_penalty;
  double distance;
  double best_dis;
  char *move;
  list_element *list;
  list_element element;

  // set ww_list if possible
  *wake_list = (ww_list *) malloc (sizeof (ww_list));
  if (*wake_list == NULL)
    {
      printf ("No mem for ww_list struct\n");
      exit (-2);
    }
  (*wake_list)->counter = 0;
  (*wake_list)->data = data;
  (*wake_list)->previous = NULL;
  // ww_data_length = sizeof [(char x_pos)+(char y_pos)+(char previous_x_pos)+(char  previous_y_pos)+(char x_break)+(double cost)+((A.y_size-1)*4Bit next_pos)]

  (*wake_list)->ww_data_length = 5 + sizeof (double) + (A.y_size / 2);
  (*wake_list)->x = 0;
  (*wake_list)->y = 0;
  (*wake_list)->y_size = A.y_size;
  (*wake_list)->beamsize = B.beamsize;

  M = A.x_size;
  N = M + 2;

  dis = malloc (sizeof (double) * M * N);
  if (dis == NULL)
    return -1;
  move = malloc (sizeof (char) * M * N);
  if (move == NULL)
    {
      free (dis);
      return -1;
    }

  /*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG*BUG
  m = 0;
  n = 3;
  while (n < N)
    {
      dis[m + (0 * M)] = DBL_MAX;
      dis[m + (1 * M)] = DBL_MAX;
      dis[m + (n * M)] = DBL_MAX;
      dis[m + ((n + 1) * M)] = DBL_MAX;
      n += 2;
      m++;
    }
  */

  for(m=0;m<M;m++)
    for(n=0;n<N;n++)
      dis[m + (n * M)] = DBL_MAX;    


//create field of all possible warpedwakes ****************************

  dis[0 + (2 * M)] = pixel_distance (0, 0, 0, 0, 0, 0, A, B);

  n_end = 3;
  n_start = 2;
  for (m = 0; m < (M - 1); m++)
    {
      n_end += 2;
      if (n_end > N)
	{
	  n_start += n_end - N;
	  n_end = N;
	}
      for (n = n_start; n < n_end; n++)
	{
	  temp_penalty = pixel_distance (0, m + 1, 0, n - 2, 0, 1, A, B);
	  temp = pixel_distance (0, m + 1, 0, n - 2, 0, 0, A, B);
	  if (dis[m + ((n - 1) * M)] + temp <=
	      dis[m + (n * M)] + temp_penalty)
	    {
	      if (dis[m + ((n - 1) * M)] + temp <=
		  dis[m + ((n - 2) * M)] + temp_penalty)
		{
		  move[m + 1 + (n * M)] = 1;
		  dis[m + 1 + (n * M)] = dis[m + ((n - 1) * M)] + temp;
		}
	      else
		{
		  move[m + 1 + (n * M)] = 2;
		  dis[m + 1 + (n * M)] = dis[m + ((n - 2) * M)]
		    + temp_penalty;
		}
	    }
	  else
	    {
	      if (dis[m + (n * M)] + temp_penalty <=
		  dis[m + ((n - 2) * M)] + temp_penalty)
		{
		  move[m + 1 + (n * M)] = 0;
		  dis[m + 1 + (n * M)] = dis[m + (n * M)] + temp_penalty;
		}
	      else
		{
		  move[m + 1 + (n * M)] = 2;
		  dis[m + 1 + (n * M)] = dis[m + ((n - 2) * M)]
		    + temp_penalty;
		}
	    }
	}
    }
//*********************************************************************

  n = N - 1;
  m = M - 1;
  distance = dis[(M * N) - 1];
  best_dis = distance;
  list = init_list (B.beamsize * M * 2);
//   printf("dis:%f\n",distance);
  ww_pos = 0;
  data[0] = 0;
  data[1] = (char) n - 2;
  data[2] = 0;
  data[3] = 0;
  data[4] = 0;

  *((double *) &data[5]) = distance;
//clear the last 4Bit if A.y_size-1 is odd
  data[(*wake_list)->ww_data_length - 1] = 0;
  while (1)
    {
      while (m > 0)
	{
	  write_y_ww (M - m - 1,
		      data + (ww_pos * (*wake_list)->ww_data_length),
		      move[m + (M * n)]);
	  write_x_ww (M - m - 1,
		      data + (ww_pos * (*wake_list)->ww_data_length), 1);
//   printf("%d-",move[m+(M*n)]);
	  element.ww_pos = ww_pos;
	  element.m_pos = m - 1;
	  element.last_move = (move[m + (M * n)] + 1) % 3;
	  element.n_pos = n - element.last_move;
	  element.distance =
	    distance + (dis[m - 1 + (M * element.n_pos)] -
			dis[m - 1 + (M * (n - move[m + (M * n)]))]) -
	    pixel_distance (0, m, 0, n - 2, 0, (move[m + (M * n)] + 1) % 2, A,
			    B) + pixel_distance (0, m, 0, n - 2, 0,
						 (element.last_move + 1) % 2,
						 A, B);
	  push_list (list, element);

	  element.ww_pos = ww_pos;
	  element.m_pos = m - 1;
	  element.last_move = (move[m + (M * n)] + 2) % 3;
	  element.n_pos = n - element.last_move;
	  element.distance =
	    distance + (dis[m - 1 + (M * element.n_pos)] -
			dis[m - 1 + (M * (n - move[m + (M * n)]))]) -
	    pixel_distance (0, m, 0, n - 2, 0, (move[m + (M * n)] + 1) % 2, A,
			    B) + pixel_distance (0, m, 0, n - 2, 0,
						 (element.last_move + 1) % 2,
						 A, B);
	  push_list (list, element);

	  n = n - move[m + (M * n)];
	  m--;
	}
      ww_pos++;
      if (ww_pos == B.beamsize)
	break;

      element = pop_list (list);
      if (element.distance >= DBL_MAX)
	break;
      if (B.dis_limit!= -1 && B.dis_limit < element.distance)
	{
	  //	  printf("*");
	  break;
	}
      for (m = 0; m < (*wake_list)->ww_data_length; m++)
	data[m + ((ww_pos) * (*wake_list)->ww_data_length)]
	  = data[m + (element.ww_pos * (*wake_list)->ww_data_length)];

//    printf("\n%d(%d,%d)%f\n",element.ww_pos,element.n_pos,element.m_pos,element.distance);
      distance = element.distance;
      *((double *) &data[5 + ((ww_pos) * (*wake_list)->ww_data_length)]) =
	distance;
      m = element.m_pos;
      n = element.n_pos;
      write_y_ww (M - m - 2, data + (ww_pos * (*wake_list)->ww_data_length),
		  element.last_move);
    }
  (*wake_list)->counter = ww_pos;
  free_list (list);

  free (dis);
  free (move);
  return (best_dis);
}

int
get_ww_position (ww_list * list, char *temp_data)
{
  int x;
  int counter;
  int overwrite;

  counter = 0;
  /*check if ww is in the list */
  while (counter < list->counter)
    {
      x = 0;
      overwrite = 1;
      while (x < list->ww_data_length && overwrite == 1)
	{
	  if (list->data[(counter * list->ww_data_length) + x] !=
	      temp_data[x])
	    overwrite = 0;
	  x++;
	  if (x == 2)
	    {
	      /*because there is no break position for list->y=0 */
	      if (list->y == 0)
		x = 13;
	      else
		x = 4;
	    }
	  if (x == 5)
	    x = 13;
	}
      if (overwrite == 1)
	return counter;
      counter++;
    }
  return -1;
}

int
insert_ww (ww_list * list, char *temp_data)
{
  int x;
  int counter;
  int temp_counter;
  double temp_distance;
  int overwrite;

  overwrite = 0;
  counter = 0;
  temp_counter = -1;
  temp_distance = *((double *) &temp_data[5]);
  /*check if ww is in the list */
  while (counter < list->counter && overwrite == 0)
    {
      x = 0;
      overwrite = 1;
      while (x < list->ww_data_length && overwrite == 1)
	{
	  if (list->data[(counter * list->ww_data_length) + x] !=
	      temp_data[x])
	    overwrite = 0;
	  x++;
	  if (x == 2)
	    {
	      /*because there is no break position for list->y=0 */
	      if (list->y == 0)
		x = 13;
	      else
		x = 4;
	    }
	  if (x == 5)
	    x = 13;
	}
      if (temp_distance <
	  *((double *) &list->data[(counter * list->ww_data_length) + 5]))
	{
	  temp_distance =
	    *((double *) &list->data[(counter * list->ww_data_length) + 5]);
	  temp_counter = counter;
	}
      counter++;
    }
  if (overwrite == 1)
    {
      /*overwrite ww */
      counter--;
      /*check which ww is better */
      if (*((double *) &list->data[(counter * list->ww_data_length) + 5]) >
	  *((double *) &temp_data[5]))
	{
	  *((double *) &list->data[(counter * list->ww_data_length) + 5]) =
	    *((double *) &temp_data[5]);
	  list->data[(counter * list->ww_data_length) + 2] = temp_data[2];
	  list->data[(counter * list->ww_data_length) + 3] = temp_data[3];
	  list->data[(counter * list->ww_data_length) + 4] = temp_data[4];
	}
    }
  else
    {
      if (list->counter < list->beamsize)
	{
	  /*insert new ww */
	  for (x = 0; x < list->ww_data_length; x++)
	    list->data[(list->counter * list->ww_data_length) + x] =
	      temp_data[x];

	  list->counter++;
	}
      else
	{
	  if (temp_counter != -1)
	    {
	      for (x = 0; x < list->ww_data_length; x++)
		list->data[(temp_counter * list->ww_data_length) + x] =
		  temp_data[x];
	    }
	}
    }
  return (counter);
}

int
beam_search (ww_list * list)
{
  double dis_min;
  double dis_max;
  int x, y, z;
  int *his;
  int his_length;
  double his_dis;
  int beamcounter;
  int his_break;

  if (list->counter <= list->beamsize)
    return 0;

  dis_min = DBL_MAX;
  dis_max = 0;

  his_length = list->beamsize * 50;

  his = malloc (sizeof (int) * (his_length + 1));
  if (his == NULL)
    {
      printf ("not enough memory for his\n");
      exit (-1);
    }
  for (x = 0; x <= his_length; x++)
    his[x] = 0;

  for (x = 0; x < list->counter; x++)
    {
      if (dis_min > *((double *) &list->data[(x * list->ww_data_length) + 5]))
	dis_min = *((double *) &list->data[(x * list->ww_data_length) + 5]);
      if (dis_max < *((double *) &list->data[(x * list->ww_data_length) + 5]))
	dis_max = *((double *) &list->data[(x * list->ww_data_length) + 5]);
    }
  if (dis_min == dis_max)
    {
//      printf("dis_min=dis_max\n");
      free (his);
      return 0;
    }
  his_dis = (dis_max - dis_min) / his_length;

  for (x = 0; x < list->counter; x++)
    his[((int)
	 ((*((double *) &list->data[(x * list->ww_data_length) + 5]) -
	   dis_min) / his_dis))]++;

//  printf("min:%f max:%f his_dis:%f\n",dis_min,dis_max,his_dis);

  beamcounter = his[0];		//how many ww will be saved
  his_break = 1;
  while (beamcounter + his[his_break] <= list->beamsize)
    {
      beamcounter += his[his_break];
      his_break++;
    }

//  printf("beam:%d his_break:%d\n",temp,his_break);

  x = 0;
  y = list->counter - 1;
  while (1)
    {
      while (x < beamcounter
	     && his_break >
	     ((int) ((*((double *) &list->data[(x * list->ww_data_length) +
					       5]) - dis_min) / his_dis)))
	x++;

      if (x >= beamcounter)
	break;

      // this should not happen...
      if (y < beamcounter)
	{
	  printf ("Something's wrong here!!!\n");
	  exit (0);
	}

      while (his_break <=
	     ((int) ((*((double *) &list->data[(y * list->ww_data_length) +
					       5]) - dis_min) / his_dis)))
	y--;

      for (z = 0; z < list->ww_data_length; z++)
	list->data[(x * list->ww_data_length) + z]
	  = list->data[(y * list->ww_data_length) + z];
      x++;
      y--;
    }
  list->counter = beamcounter;

  free (his);
  return 0;
}

int
first_ww_distance (ww_list * list, int data_pos, IMG A, IMG B)
{
  int y_pos;
  int data_ww_pos;
  int ww_y_move;

  y_pos = list->data[data_pos + 1];

  *((double *) &list->data[data_pos + 5]) =
    pixel_distance (0, A.y_size - 1, 0, y_pos, 0, 0, A, B);
  data_ww_pos = 0;
  while (data_ww_pos < A.y_size - 1)
    {
      ww_y_move = read_y_ww (data_ww_pos, list->data + data_pos);
      y_pos -= ww_y_move;
      *((double *) &list->data[data_pos + 5]) +=
	pixel_distance (0, A.y_size - 2 - data_ww_pos, 0, y_pos,
			0, abs (ww_y_move - 1), A, B);
      data_ww_pos++;
    }
  return (0);
}

void init_ww_list_distance (ww_list * list, IMG A, IMG B)
{
  int counter;
  int y_pos;
  int data_ww_pos;
  int ww_y_move;

  for (counter = 0; counter < list->counter; counter++)
    {
      y_pos = list->data[(counter * list->ww_data_length) + 1];

      *((double *) &list->data[(counter * list->ww_data_length) + 5]) =
	pixel_distance (0, A.y_size - 1, 0, y_pos, 0, 0, A, B);
      data_ww_pos = 0;
      while (data_ww_pos < A.y_size - 1)
	{
	  ww_y_move = read_y_ww (data_ww_pos,
				 list->data +
				 (counter * list->ww_data_length));
	  y_pos -= ww_y_move;
	  *((double *) &list->data[(counter * list->ww_data_length) + 5]) +=
	    pixel_distance (0, A.y_size - 2 - data_ww_pos, 0, y_pos,
			    0, abs (ww_y_move - 1), A, B);
	  data_ww_pos++;
	}
    }
  beam_search (list);

}

inline int
ww_compare (char *temp_data, int position, ww_list * list)
{
  int x;
  int equal = 1;

  for (x = 0; x < list->ww_data_length; x++)
    {
      if (x < 6)
	switch (x)
	  {
	  case 2:
	    /*because there is no break position for list->y=0 */
	    if (list->y == 0)
	      x = 13;
	    else
	      x = 4;
	    break;
	  case 5:
	    x = 13;
	    break;
	  }

      if (temp_data[x] != list->data[(position * list->ww_data_length) + x])
	{
	  equal = 0;
	  x = list->ww_data_length;
	}
    }
  return equal;
}

inline int
hash_index (ww_list * list, char *data, hash_list * hash)
{
  long unsigned int index;
  int x;
  int x_start = list->y_size - 2;
  index = 0;

  for (x = x_start; x >= 0; x--)
    index += read_x_ww (x, data) * read_y_ww (x, data) * POW9[x_start - x];
//  index+=data[4]*pow(9,15); //break-start of WW
  index = index % hash->length;
// index = (*((unsigned int *) (&data[list->ww_data_length-8]))) % hash->length;
  return index;
}

int
hash_init_POW9 ()
{
  int x;

  for (x = 0; x < 32; x++)
    POW9[x] = pow (9, x);

  return (0);
}

int
hash_init (ww_list * list, hash_list ** hash)
{
  int x;
  int prim[17] =
    { 101, 151, 211, 311, 421, 631, 839, 1223, 1693, 3347, 6607, 13001, 26203,
    52009, 100003, 227873, 420929
  };
  *hash = malloc (sizeof (hash_list));

  if (*hash == NULL)
    {
      printf ("not enough memory for hash_list\n");
      exit (-1);
    }

  for (x = 0; x < 13; x++)
    if (prim[x] > (list->previous)->counter)
      break;
  if (prim[x] < (list->previous)->counter)
    (*hash)->length = (list->previous)->counter;
  else
    (*hash)->length = prim[x];

  (*hash)->data = malloc (sizeof (hash_el *) * (*hash)->length);
  if ((*hash)->data == NULL)
    {
      printf ("not enough memory for hash_list\n");
      exit (-1);
    }
  for (x = 0; x < (*hash)->length; x++)
    (*hash)->data[x] = NULL;
  return 0;
}

int
hash_free (hash_list * hash)
{
  int x;
  hash_el *hash_ptr;
  hash_el *hash_temp_ptr;

  for (x = 0; x < hash->length; x++)
    {
      hash_ptr = hash->data[x];
      while (hash_ptr != NULL)
	{
	  hash_temp_ptr = hash_ptr;
	  hash_ptr = hash_ptr->next;
	  free (hash_temp_ptr);
	}
    }
  free (hash->data);
  free (hash);

  return 0;
}

int
hash_insert_ww (ww_list * list, char *temp_data, hash_list * hash)
{
  int x;
  hash_el **hash_ptr;

  hash_ptr = &(hash->data[hash_index (list, temp_data, hash)]);

  x = 0;
  while ((*hash_ptr) != NULL)
    {
      if (ww_compare (temp_data, (*hash_ptr)->position, list))
	break;
      hash_ptr = &((*hash_ptr)->next);
      x++;
    }
//  if(x>10)
//    printf("collision:%d\n",x);

  if ((*hash_ptr) == NULL)
    {
      (*hash_ptr) = malloc (sizeof (hash_el));
      if ((*hash_ptr) == NULL)
	{
	  printf ("not bla");
	  exit (-1);
	}
      (*hash_ptr)->next = NULL;
      (*hash_ptr)->position = list->counter;

      for (x = 0; x < list->ww_data_length; x++)
	list->data[(list->counter * list->ww_data_length) + x] = temp_data[x];
      list->counter++;
    }
  else
    {
      if (*
	  ((double *) &list->
	   data[(*hash_ptr)->position * list->ww_data_length + 5]) >
	  *((double *) &temp_data[5]))
	{
	  *((double *) &list->
	    data[(*hash_ptr)->position * list->ww_data_length + 5]) =
	    *((double *) &temp_data[5]);
	  list->data[((*hash_ptr)->position * list->ww_data_length) + 2] =
	    temp_data[2];
	  list->data[((*hash_ptr)->position * list->ww_data_length) + 3] =
	    temp_data[3];
	  list->data[((*hash_ptr)->position * list->ww_data_length) + 4] =
	    temp_data[4];
	}
    }
  return 0;
}

double
next_ww_list (ww_list ** list, IMG A, IMG B)
{
  ww_list *new_list;
  int x;
  int data_ww_pos;
  int data_pos;
  int B_x_pos;
  int B_y_pos;
  int ww_x_move;
  int ww_y_move;
  char *temp_data;
  hash_list *hash;
  double best_dis=DBL_MAX;
  double dis;

  new_list = (ww_list *) malloc (sizeof (ww_list));
  if (new_list == 0)
    {
      printf ("No mem for ww_list struct\n");
      exit (-1);
    }
  temp_data = malloc (sizeof (char) * (*list)->ww_data_length);
  if (temp_data == 0)
    {
      printf ("No mem for temp_data\n");
      exit (-1);
    }
  new_list->ww_data_length = (*list)->ww_data_length;
  new_list->y_size = (*list)->y_size;
  new_list->data =
    (*list)->data + ((*list)->counter * (*list)->ww_data_length);
  new_list->counter = 0;
  new_list->previous = *list;
  new_list->y = (*list)->y - 1;
  new_list->x = (*list)->x;
  new_list->beamsize = (*list)->beamsize;
  if (new_list->y < 0)
    {
      new_list->y = A.y_size - 1;
      new_list->x++;
    }
  hash_init (new_list, &hash);
  data_pos = 0;
  for (x = 0; x < (*list)->counter; x++)
    {

      /*calculate the last pixel coordinates of ww */
      B_x_pos = (*list)->data[data_pos];
      B_y_pos = (*list)->data[data_pos + 1];

      data_ww_pos = 0;
      while (data_ww_pos < A.y_size - 1)
	{
	  if (data_ww_pos == (*list)->y - 1)	/*find break position in ww */
	    {
	      B_x_pos = (*list)->data[data_pos + 4];
	      B_y_pos = B.y_size - 1;
	    }
	  else
	    {
	      B_x_pos +=
		read_x_ww (data_ww_pos, (*list)->data + data_pos) - 1;

	      B_y_pos -= read_y_ww (data_ww_pos, (*list)->data + data_pos);;
	    }
	  if (data_ww_pos == 0)
	    {
	      temp_data[0] = B_x_pos;
	      temp_data[1] = B_y_pos;
	      temp_data[2] = (*list)->data[data_pos];
	      temp_data[3] = (*list)->data[data_pos + 1];
	      temp_data[4] = (*list)->data[data_pos + 4];
	      temp_data[(*list)->ww_data_length - 1] = 0;
	    }
	  data_ww_pos++;
	}

      for (data_ww_pos = 0; data_ww_pos < A.y_size - 2; data_ww_pos++)
	{
	  write_x_ww (data_ww_pos, temp_data,
		      read_x_ww (data_ww_pos + 1, (*list)->data + data_pos));
	  write_y_ww (data_ww_pos, temp_data,
		      read_y_ww (data_ww_pos + 1, (*list)->data + data_pos));
	}

      /*check for new break position in ww */
      if (new_list->y == (A.y_size - 1))
	{
	  /*new break position in ww */
	  write_x_ww (A.y_size - 2, temp_data, 0);
	  write_y_ww (A.y_size - 2, temp_data, 0);
	  ww_x_move = 0;
	  for (B_x_pos = (*list)->data[data_pos];
	       B_x_pos < (*list)->data[data_pos] + 3; B_x_pos++)
	    {
	      temp_data[4] = (char) B_x_pos;
	      /*check if the temp_ww(temp_data) is possible */

	      if ((B_x_pos < B.x_size) &&
		  ((B_x_pos + ((A.x_size - 1 - new_list->x) * 2)) >=
		   (B.x_size - 1)) && (B_y_pos >= 0)
		  && ((B_y_pos - ((new_list->y) * 2)) <= 0))
		{
		  *((double *) &temp_data[5]) =
		    *((double *) &(*list)->data[data_pos + 5]) +
		    pixel_distance (new_list->x, new_list->y, B_x_pos,
				    B.y_size - 1, abs (ww_x_move - 1), 0, A,
				    B);

		  //insert_ww (new_list, temp_data);
		  dis = *((double *) &temp_data[5]);
		  if (dis < DBL_MAX )
		    {
		      if(!(B.dis_limit!= -1 && B.dis_limit < dis))
			{
		      hash_insert_ww (new_list, temp_data, hash);
		      if(best_dis > dis)
			best_dis=dis;
			}
		      // else{printf(".");}
		    }
		}
	      ww_x_move++;
	    }
	}
      else
	{
	  /*coordinates for the next pixel in ww */
	  for (ww_x_move = 0; ww_x_move < 3; ww_x_move++)
	    {
	      for (ww_y_move = 0; ww_y_move < 3; ww_y_move++)
		{

		  if (((B_x_pos - 1 + ww_x_move) >= (*list)->data[data_pos])
		      && ((B_x_pos - 1 + ww_x_move) <=
			  ((*list)->data[data_pos] + 2))
		      && ((B_y_pos - ww_y_move) <=
			  ((*list)->data[data_pos + 1] + 1))
		      && ((B_y_pos - ww_y_move) >=
			  ((*list)->data[data_pos + 1] - 1)))
		    {
		      /*if temp_ww is a possible ww then insert temp_ww in new_list */
		      if (((B_x_pos - 1 + ww_x_move) < B.x_size) &&
			  ((B_x_pos
			    - 1 + ww_x_move +
			    ((A.x_size - 1 - new_list->x) * 2)) >=
			   (B.x_size - 1)) && ((B_y_pos - ww_y_move) >= 0)
			  && ((B_y_pos - ww_y_move - ((new_list->y) * 2)) <=
			      0))
			{
			  write_x_ww (A.y_size - 2, temp_data, ww_x_move);
			  write_y_ww (A.y_size - 2, temp_data, ww_y_move);
			  *((double *) &temp_data[5]) =
			    *((double *) &(*list)->data[data_pos + 5]) +
			    pixel_distance (new_list->x, new_list->y,
					    B_x_pos - 1 + ww_x_move,
					    B_y_pos - ww_y_move,
					    abs (B_x_pos - 1 + ww_x_move -
						 (*list)->data[data_pos] -
						 1) + abs (B_y_pos -
							   ww_y_move -
							   (*list)->
							   data[data_pos +
								1]),
					    abs (ww_x_move - 1) +
					    abs (ww_y_move - 1), A, B);
		
			  //insert_ww (new_list, temp_data);
			  dis = *((double *) &temp_data[5]);
			  if (dis < DBL_MAX)
			    {
			      if(!(B.dis_limit!= -1 && B.dis_limit < dis))
				{
				  hash_insert_ww (new_list, temp_data, hash);
				  if(best_dis > dis)
				    best_dis=dis;
				}
			      // else{printf(".");}
			    }
			}
		    }
		}
	    }
	}

      data_pos = data_pos + (*list)->ww_data_length;
    }
  free (temp_data);
  hash_free (hash);
  *list = new_list;

  data_pos = (*list)->counter + 1;
  while ((*list)->counter > (*list)->beamsize && (*list)->counter != data_pos)
    {
      data_pos = (*list)->counter;
      beam_search (*list);
      //   printf("after beamsize: %d WW\n",(*list)->counter);
    }
  //printf("\n");
  if ((*list)->counter > (*list)->beamsize)
    (*list)->counter = (*list)->beamsize;

  return (best_dis);
}

void print_ww (ww_list * list, int data_pos, IMG A, IMG B)
{
  int x_pos;
  int y_pos;
  int data_ww_pos;

  x_pos = (char) list->data[data_pos + 0];
  y_pos = (char) list->data[data_pos + 1];

  printf ("Prev.:(%d,%d)", list->data[data_pos + 2],
	  list->data[data_pos + 3]);
  printf ("[(%d,%d)", x_pos, y_pos);

  data_ww_pos = 0;
  while (data_ww_pos < A.y_size - 1)
    {
      if (data_ww_pos == list->y - 1)	/*find break position in ww */
	{
	  /* break */
	  x_pos = list->data[data_pos + 4];
	  y_pos = B.y_size - 1;
	}
      else
	{
	  x_pos = x_pos + read_x_ww (data_ww_pos, list->data + data_pos) - 1;
	  y_pos = y_pos - read_y_ww (data_ww_pos, list->data + data_pos);
	}

      printf (",(%d,%d)", x_pos, y_pos);
      data_ww_pos++;
    }
  printf ("brk:%d", list->data[data_pos + 4]);
  printf ("] D:%f", *((double *) &list->data[data_pos + 5]));
  printf ("\n");
}

void print_ww_list (ww_list * list, IMG A, IMG B)
{
  int counter;
  printf ("List of warped wakes from stage (%d,%d)\n", list->x, list->y);
  for (counter = 0; counter < list->counter; counter++)
    {
      printf ("ww:%d ", counter + 1);
      print_ww (list, counter * list->ww_data_length, A, B);
    }
}

int
initialize_wimage_data (int **wimage_data, int image_counter, int dimension_A,
			int dimension_B)
{
  int x;

  if ((*wimage_data =
       malloc ((dimension_B +
		18) * dimension_A * image_counter * sizeof (int))) == NULL)
    return (0);
  for (x = 0; x < (dimension_B + 18) * dimension_A * image_counter; x++)
    (*wimage_data)[x] = 0;

  return (1);
}

int
increase_wimage_data (int *wimage_data, ww_list * list, int best_ww)
{
  char *ww_temp_data;
  int *warping_data;
  int data_ww_pos;
  int A_x_pos;
  int A_y_pos;
  int B_x_pos;
  int B_y_pos;
  int A_x_size;
  int A_y_size;
  int B_x_size;
  int B_y_size;
  int x_diff;
  int y_diff;

  A_x_size = list->x + 1;
  A_y_size = list->x + 1;	//only

  warping_data = malloc (sizeof (int) * A_x_size * A_y_size * 2);
  if (warping_data == 0)
    {
      printf ("No mem for warping_data\n");
      exit (-2);
    }

  ww_temp_data = malloc (sizeof (char) * list->ww_data_length);
  if (ww_temp_data == 0)
    {
      free (warping_data);
      printf ("No mem for ww_temp_data\n");
      exit (-2);
    }
  B_x_size = list->data[0] + 1;
  B_y_size = list->data[1] + 1;
  A_x_pos = A_x_size - 1;
  A_y_pos = A_y_size - 1;
  B_x_pos = list->data[(best_ww * list->ww_data_length) + 0];
  B_y_pos = list->data[(best_ww * list->ww_data_length) + 1];
  warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2)] = B_x_pos;
  warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2) + 1] = B_y_pos;

  for (data_ww_pos = 0; data_ww_pos < A_y_size - 1; data_ww_pos++)
    {
      B_y_pos =
	B_y_pos - read_y_ww (data_ww_pos,
			     &list->data[best_ww * list->ww_data_length]);
      B_x_pos =
	B_x_pos + read_x_ww (data_ww_pos,
			     &list->data[best_ww * list->ww_data_length]) - 1;
      A_y_pos--;
      warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2)] = B_x_pos;
      warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2) + 1] = B_y_pos;
    }
  while (list->previous != 0)
    {
      ww_temp_data[0] = list->data[(best_ww * list->ww_data_length) + 2];
      ww_temp_data[1] = list->data[(best_ww * list->ww_data_length) + 3];
      ww_temp_data[4] = list->data[(best_ww * list->ww_data_length) + 4];
      ww_temp_data[list->ww_data_length - 1] = 0;
      for (data_ww_pos = 0; data_ww_pos < A_y_size - 2; data_ww_pos++)
	{
	  write_x_ww (data_ww_pos + 1, ww_temp_data,
		      read_x_ww (data_ww_pos,
				 &list->data[best_ww *
					     list->ww_data_length]));
	  write_y_ww (data_ww_pos + 1, ww_temp_data,
		      read_y_ww (data_ww_pos,
				 &list->data[best_ww *
					     list->ww_data_length]));
	}
      if (list->y == 0)
	{
	  write_x_ww (0, ww_temp_data, 0);
	  write_y_ww (0, ww_temp_data, 0);
	}
      else
	{
	  write_x_ww (0, ww_temp_data,
		      (list->data[(best_ww * list->ww_data_length)] + 1 -
		       list->data[(best_ww * list->ww_data_length) + 2]));
	  write_y_ww (0, ww_temp_data,
		      (list->data[(best_ww * list->ww_data_length) + 3] -
		       list->data[(best_ww * list->ww_data_length) + 1]));
	}

      list = list->previous;
      best_ww = get_ww_position (list, ww_temp_data);
      if (list->y == 0)
	{
	  A_x_pos--;
	  A_y_pos = A_y_size - 1;
	  B_x_pos = list->data[(best_ww * list->ww_data_length) + 0];
	  B_y_pos = list->data[(best_ww * list->ww_data_length) + 1];
	  warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2)] = B_x_pos;
	  warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2) + 1] = B_y_pos;
	  for (data_ww_pos = 0; data_ww_pos < A_y_size - 1; data_ww_pos++)
	    {
	      B_y_pos -=
		read_y_ww (data_ww_pos,
			   &list->data[best_ww * list->ww_data_length]);
	      B_x_pos +=
		read_x_ww (data_ww_pos,
			   &list->data[best_ww * list->ww_data_length]) - 1;
	      A_y_pos--;
	      warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2)] = B_x_pos;
	      warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2) + 1] =
		B_y_pos;
	    }
	}
    }

  for (A_x_pos = 0; A_x_pos < A_x_size; A_x_pos++)
    for (A_y_pos = 0; A_y_pos < A_y_size; A_y_pos++)
      {
	B_x_pos = warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2)];
	B_y_pos = warping_data[(((A_y_pos * A_x_size) + A_x_pos) * 2) + 1];

	wimage_data[((B_y_size * B_x_size) + 18) * ((A_y_pos * A_x_size) +
						    A_x_pos) +
		    (B_y_pos * B_x_size) + B_x_pos]++;
	if (A_x_pos != 0)
	  {
	    x_diff =
	      B_x_pos -
	      warping_data[(((A_y_pos * A_x_size) + A_x_pos - 1) * 2)];
	    y_diff =
	      B_y_pos -
	      warping_data[(((A_y_pos * A_x_size) + A_x_pos - 1) * 2) + 1] +
	      1;
	  }
	else
	  {
	    x_diff = 1;
	    y_diff = 1;
	  }
	wimage_data[((B_y_size * B_x_size) + 18) * ((A_y_pos * A_x_size) + A_x_pos + 1) - 18 + (y_diff * 3) + x_diff]++;	//neighbour
	if (A_y_pos != 0)
	  {
	    x_diff =
	      B_x_pos -
	      warping_data[((((A_y_pos - 1) * A_x_size) + A_x_pos) * 2)] + 1;
	    y_diff =
	      B_y_pos -
	      warping_data[((((A_y_pos - 1) * A_x_size) + A_x_pos) * 2) + 1];
	  }
	else
	  {
	    x_diff = 1;
	    y_diff = 1;
	  }
	wimage_data[((B_y_size * B_x_size) + 18) * ((A_y_pos * A_x_size) + A_x_pos + 1) - 9 + (y_diff * 3) + x_diff]++;	//previous
      }
  free (ww_temp_data);
  free (warping_data);
  return (0);
}

int
print_wimage_data (int *wimage_data, int x_size, int y_size)
{
  int x, y, i, j, temp;

  printf ("warping\n");
  for (y = 0; y < y_size; y++)
    {
      for (j = -1; j < 2; j++)
	{
	  for (x = 0; x < x_size; x++)
	    {
	      for (i = -1; i < 2; i++)
		{
		  if ((x + i) >= 0 && (x + i) < x_size && (y + j) >= 0
		      && (y + j) < y_size)
		    temp =
		      wimage_data[(((x_size * y_size) + 18) * ((y * x_size) +
							       x)) + ((y +
								       j) *
								      x_size)
				  + x + i];
		  else
		    temp = 0;
		  printf ("%3d,", temp);
		}
	      printf (" ");
	    }
	  printf ("\n");
	}
      printf ("\n");
    }

  printf ("neighbour\n");
  for (y = 0; y < y_size; y++)
    {
      for (i = 0; i < 3; i++)
	{
	  for (x = 0; x < x_size; x++)
	    {
	      printf ("%3d,%3d,%3d, ",
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) - 18],
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) + 1 - 18],
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) + 2 - 18]);
	    }
	  printf ("\n");
	}
      printf ("\n");
    }

  printf ("previous\n");
  for (y = 0; y < y_size; y++)
    {
      for (i = 0; i < 3; i++)
	{
	  for (x = 0; x < x_size; x++)
	    {
	      printf ("%3d,%3d,%3d ",
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) - 9],
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) + 1 - 9],
		      wimage_data[((x_size * y_size) + 18) * ((y * x_size) +
							      x + 1) +
				  (i * 3) + 2 - 9]);
	    }
	  printf ("\n");
	}
      printf ("\n");
    }
  return (0);
}

void create_warping_info (ww_list * list, int best_ww, int *winfo, IMG A, IMG B)
{
  char *temp_data;
  int data_ww_pos;
  int A_x_pos;
  int A_y_pos;
  int B_x_pos;
  int B_y_pos;

  temp_data = malloc (sizeof (char) * list->ww_data_length);
  if (temp_data == 0)
    {
      printf ("No mem for temp_data\n");
      exit (-2);
    }

  A_x_pos = A.x_size - 1;
  A_y_pos = A.y_size - 1;
  B_x_pos = list->data[(best_ww * list->ww_data_length) + 0];
  B_y_pos = list->data[(best_ww * list->ww_data_length) + 1];
  winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2)] = B_x_pos;
  winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2) + 1] = B_y_pos;

  for (data_ww_pos = 0; data_ww_pos < A.y_size - 1; data_ww_pos++)
    {
      B_y_pos =
	B_y_pos - read_y_ww (data_ww_pos,
			     &list->data[best_ww * list->ww_data_length]);
      B_x_pos =
	B_x_pos + read_x_ww (data_ww_pos,
			     &list->data[best_ww * list->ww_data_length]) - 1;
      A_y_pos--;
      winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2)] = B_x_pos;
      winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2) + 1] = B_y_pos;
    }
  while (list->previous != 0)
    {
      temp_data[0] = list->data[(best_ww * list->ww_data_length) + 2];
      temp_data[1] = list->data[(best_ww * list->ww_data_length) + 3];
      temp_data[4] = list->data[(best_ww * list->ww_data_length) + 4];
      temp_data[list->ww_data_length - 1] = 0;
      for (data_ww_pos = 0; data_ww_pos < A.y_size - 2; data_ww_pos++)
	{
	  write_x_ww (data_ww_pos + 1, temp_data,
		      read_x_ww (data_ww_pos,
				 &list->data[best_ww *
					     list->ww_data_length]));
	  write_y_ww (data_ww_pos + 1, temp_data,
		      read_y_ww (data_ww_pos,
				 &list->data[best_ww *
					     list->ww_data_length]));
	}
      if (list->y == 0)
	{
	  write_x_ww (0, temp_data, 0);
	  write_y_ww (0, temp_data, 0);
	}
      else
	{
	  write_x_ww (0, temp_data,
		      (list->data[(best_ww * list->ww_data_length)] + 1 -
		       list->data[(best_ww * list->ww_data_length) + 2]));
	  write_y_ww (0, temp_data,
		      (list->data[(best_ww * list->ww_data_length) + 3] -
		       list->data[(best_ww * list->ww_data_length) + 1]));
	}

      list = list->previous;
      best_ww = get_ww_position (list, temp_data);
      if (list->y == 0)
	{
	  A_x_pos--;
	  A_y_pos = A.y_size - 1;
	  B_x_pos = list->data[(best_ww * list->ww_data_length) + 0];
	  B_y_pos = list->data[(best_ww * list->ww_data_length) + 1];
	  winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2)] = B_x_pos;
	  winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2) + 1] = B_y_pos;
	  for (data_ww_pos = 0; data_ww_pos < A.y_size - 1; data_ww_pos++)
	    {
	      B_y_pos -=
		read_y_ww (data_ww_pos,
			   &list->data[best_ww * list->ww_data_length]);
	      B_x_pos +=
		read_x_ww (data_ww_pos,
			   &list->data[best_ww * list->ww_data_length]) - 1;
	      A_y_pos--;
	      winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2)] = B_x_pos;
	      winfo[(((A_y_pos * A.x_size) + A_x_pos) * 2) + 1] = B_y_pos;
	    }
	}
    }
  free (temp_data);
}

int
free_ww_list (ww_list * list)
{
  ww_list *temp_list;

  while (list != NULL)
    {
      temp_list = list;
      list = list->previous;
      free (temp_list);
    }
  return 0;
}

double
get_best_ww_list_distance (ww_list * list)
{
  int counter;
  double best_distance;

  best_distance = DBL_MAX;
  for (counter = 0; counter < list->counter; counter++)
    {
      if (best_distance >
	  *((double *) &(list->data[(counter * list->ww_data_length) + 5])))
	best_distance =
	  *((double *) &(list->data[(counter * list->ww_data_length) + 5]));
    }
  return (best_distance);
}

double
get_worst_ww_list_distance (ww_list * list)
{
  int counter;
  double worst_distance;

  worst_distance = 0;
  for (counter = 0; counter < list->counter; counter++)
    {
      if (worst_distance <
	  *((double *) &list->data[(counter * list->ww_data_length) + 5]))
	worst_distance =
	  *((double *) &list->data[(counter * list->ww_data_length) + 5]);
    }
  return worst_distance;
}

double
W2DW_list_distance(IMG A, IMG_list* train_img_list, double dis_limit, int *warping_info)
{
  int img_nr;
  ww_list **list;
  ww_list *temp_list;
  char **data;
  double dis,best_dis,new_dis_limit;
  int best_img_nr;
  int x;
  int ww_sum=0;

  list=malloc(sizeof(ww_list*) * train_img_list->total_nr);
  data=malloc(sizeof(char*) * train_img_list->total_nr);

  for(img_nr=0;img_nr<train_img_list->total_nr;img_nr++)
    {
      data[img_nr] =
	malloc (sizeof (char) * (5 + sizeof (double) + (A.y_size / 2)) *
		train_img_list->img[img_nr].beamsize * ((A.x_size - 1) * A.y_size + 10));
      if (data[img_nr] == NULL)
	{
	  fprintf (stderr, "not enough memory for W2DW_distance, Beamsize:%d\n",
		   train_img_list->img[img_nr].beamsize);
	  exit (-1);
	}     
    }

  hash_init_POW9 ();

  best_dis=DBL_MAX;
  for(img_nr=0;img_nr<train_img_list->total_nr;img_nr++)
    {
      train_img_list->img[img_nr].dis_limit=dis_limit;
      dis=init_ww_list (&(list[img_nr]), A, train_img_list->img[img_nr],data[img_nr]);
      ww_sum += list[img_nr]->counter;
      if(dis < best_dis)
	best_dis=dis;
   }

  for (x = 0; x < ((A.x_size - 1) * A.y_size); x++)
    {
      new_dis_limit=best_dis;
      best_dis=DBL_MAX;
      while(best_dis > (DBL_MAX/2))
	{
	  new_dis_limit += dis_limit;
	  for(img_nr=0;img_nr<train_img_list->total_nr;img_nr++)
	    {
	      train_img_list->img[img_nr].dis_limit=new_dis_limit;
	      dis=next_ww_list (&(list[img_nr]), A,train_img_list->img[img_nr]);
	      ww_sum += list[img_nr]->counter;
	      if(dis < best_dis)
		{
		  best_dis=dis;
		  best_img_nr=img_nr;
		}
	    }
	  if(best_dis > (DBL_MAX/2))
	    {
	      //	      printf("*");
	       for(img_nr=0;img_nr<train_img_list->total_nr;img_nr++)
		 {
		   temp_list=list[img_nr];
		   list[img_nr]=list[img_nr]->previous;
		   free(temp_list);
		 }
	    }
	}
    }
   for(img_nr=0;img_nr<train_img_list->total_nr;img_nr++)
     {
       
       if(list[img_nr]->counter>0)
	 {
	   best_dis=get_best_ww_list_distance(list[img_nr]);
	   fprintf(stdout,"%s %s %d %d %f\n",A.name,train_img_list->img[img_nr].name,
		   A.class,train_img_list->img[img_nr].class, best_dis);
	 }
       free(data[img_nr]);
       free_ww_list (list[img_nr]);
     }

   free(list);
   free(data);

   fprintf(stdout,"# WW_SUM: %d\n",ww_sum/((((A.x_size - 1) * A.y_size)+1)*train_img_list->total_nr));
   return best_dis;
}


double
W2DW_distance (IMG A, IMG B, int *warping_info)
{
  ww_list *list;
  char *data;
  int best_ww;
  int x;
  double distance;

  data =
    malloc (sizeof (char) * (5 + sizeof (double) + (A.y_size / 2)) *
	    B.beamsize * ((A.x_size - 1) * A.y_size + 10));
  if (data == NULL)
    {
      fprintf (stderr, "not enough memory for W2DW_distance, Beamsize:%d\n",
	       B.beamsize);
      exit (-1);
    }

  hash_init_POW9 ();

  init_ww_list (&list, A, B, data);	// init_ww_list

  for (x = 0; x < ((A.x_size - 1) * A.y_size); x++)
    next_ww_list (&list, A, B);

  // get distance of first ww of last ww_list

  distance = *((double *) &list->data[5]);
  best_ww = 0;
  for (x = 1; x < list->counter; x++)
    {
      if (distance >
	  *((double *) &list->data[(x * list->ww_data_length) + 5]))
	{
	  best_ww = x;
	  distance =
	    *((double *) &list->data[(x * list->ww_data_length) + 5]);
	}
    }
 
  if (warping_info != NULL)
    create_warping_info (list, best_ww, warping_info, A, B);

  free (data);
  free_ww_list (list);
  return distance;
}

double
dis_from_warp_image (IMG A, IMG B, int *warp_data)
{
  int x, y, warp_x, warp_y, left_x, left_y, under_x, under_y;
  double distance;

  distance = 0;
  for (y = 0; y < A.y_size; y++)
    for (x = 0; x < A.x_size; x++)
      {
	warp_x = warp_data[((y * A.x_size) + x) * 2];
	warp_y = warp_data[(((y * A.x_size) + x) * 2) + 1];
	if (x != 0)
	  {
	    left_x = warp_data[((y * A.x_size) + x - 1) * 2];
	    left_y = warp_data[(((y * A.x_size) + x - 1) * 2) + 1];
	  }
	else
	  {
	    left_x = -1;
	    left_y = warp_y;
	  }

	if (y != 0)
	  {
	    under_x = warp_data[(((y - 1) * A.x_size) + x) * 2];
	    under_y = warp_data[((((y - 1) * A.x_size) + x) * 2) + 1];
	  }
	else
	  {
	    under_x = warp_x;
	    under_y = -1;
	  }
	distance += pixel_distance (x, y, warp_x, warp_y,
				    abs (left_x + 1 - warp_x) + abs (left_y -
								     warp_y),
				    abs (under_x - warp_x) + abs (under_y +
								  1 - warp_y),
				    A, B);
      }
  return (distance);
}
